#!/usr/bin/env python
"""Semantic Protobufs are serialization agnostic, rich data types."""

import base64
from collections import abc
from collections.abc import ByteString, Iterator, Sequence
import copy
import functools
import logging
import struct
from typing import Optional, TypeVar, cast

from google.protobuf import any_pb2
from google.protobuf import wrappers_pb2
from google.protobuf import message as proto2_message
from google.protobuf import text_format
from grr_response_core import _semantic
from grr_response_core.lib import rdfvalue
from grr_response_core.lib import registry
from grr_response_core.lib import serialization
from grr_response_core.lib import type_info
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import proto2 as rdf_proto2
from grr_response_core.lib.util import precondition
from grr_response_proto import semantic_pb2


# pylint: disable=invalid-name
VarintEncode = _semantic.varint_encode
VarintReader = _semantic.varint_decode
SplitBuffer = _semantic.split_buffer
# pylint: enable=invalid-name

# These numbers identify the wire type of a protocol buffer value.
# We use the least-significant TAG_TYPE_BITS bits of the varint-encoded
# tag-and-type to store one of these WIRETYPE_* constants.
# These values must match WireType enum in google/protobuf/wire_format.h.
WIRETYPE_VARINT = 0
WIRETYPE_FIXED64 = 1
WIRETYPE_LENGTH_DELIMITED = 2

# We do not support these deprecated wire types any more. Nested protobufs are
# stored using the normal WIRETYPE_LENGTH_DELIMITED tag.
WIRETYPE_START_GROUP = 3
WIRETYPE_END_GROUP = 4

WIRETYPE_FIXED32 = 5


def SignedVarintEncode(value):
  """Encode a signed integer as a signed varint."""
  if value < 0:
    value += 1 << 64

  return VarintEncode(value)


def SignedVarintReader(buf, pos=0):
  """A signed 64 bit decoder for signed varints."""
  result, p = VarintReader(buf, pos)
  if result > 0x7FFFFFFFFFFFFFFF:
    result -= 1 << 64

  return (result, p)


def _GetOrderedEntries(data):
  """Gets entries of `RDFProtoStruct` in a well-defined order.

  Args:
    data: A raw data dictionary of `RDFProtoStruct`.

  Returns:
    Entries of the structured in a well-defined order.
  """

  # Sort struct entries by their field tag.
  def Tag(entry):
    _, wire_format, type_descriptor = entry
    if type_descriptor is not None:
      encoded_tag = type_descriptor.encoded_tag
    elif wire_format is not None:
      encoded_tag = wire_format[0]
    else:
      raise AssertionError(
          "Each entry is expected to have "
          "either a type_descriptor or wire_format."
      )

    return VarintReader(encoded_tag, 0)[0]

  return sorted(data.values(), key=Tag)


def _SerializeEntries(entries):
  """Serializes given triplets of python and wire values and a descriptor."""

  output = []
  for python_format, wire_format, type_descriptor in entries:

    if wire_format is None or (
        python_format and type_descriptor.IsDirty(python_format)
    ):
      wire_format = type_descriptor.ConvertToWireFormat(python_format)

    precondition.AssertIterableType(wire_format, bytes)
    output.extend(wire_format)

  return b"".join(output)


def ReadIntoObject(buff, index, value_obj, length=0):
  """Reads all tags until the next end group and store in the value_obj."""
  raw_data = value_obj.GetRawData()
  count = 0

  # Split the buffer into tags and wire_format representations, then collect
  # these into the raw data cache.
  for encoded_tag, encoded_length, encoded_field in SplitBuffer(
      buff, index=index, length=length
  ):

    type_info_obj = value_obj.type_infos_by_encoded_tag.get(encoded_tag)

    # Internal format to store parsed fields.
    wire_format = (encoded_tag, encoded_length, encoded_field)

    # If the tag is not found we need to skip it. Skipped fields are
    # inaccessible to this actual object, because they have no type info
    # describing them, however they are still stored in the raw data
    # representation because they will be re-serialized back. This way
    # programs which simply read protobufs and write them back do not need to
    # know all the fields, some of which were defined in a later version of
    # the application. In order to avoid having to worry about repeated fields
    # here, we just insert them into the raw data dict with a key which should
    # be unique.
    if type_info_obj is None:
      # Record an unknown field. The key is unique and ensures we do not collide
      # the dict on repeated fields of the encoded tag. Note that this field is
      # not really accessible using Get() and does not have a python format
      # representation. It will be written back using the same wire format it
      # was read with, therefore does not require a type descriptor at all.
      field_nr = VarintReader(encoded_tag, 0)[0] >> 3
      raw_data["_unknown_field_%d" % field_nr] = (None, wire_format, None)

      count += 1

    # Repeated fields are handled especially.
    elif type_info_obj.__class__ is ProtoList:
      value_obj.Get(type_info_obj.name).wrapped_list.append((None, wire_format))

    else:
      # Set the python_format as None so it gets converted lazily on access.
      raw_data[type_info_obj.name] = (None, wire_format, type_info_obj)

  value_obj.SetRawData(raw_data)


class ProtoType(type_info.TypeInfoObject):
  """A specific type descriptor for protobuf fields.

  This is an abstract class - do not instantiate directly.
  """

  # Must be overridden by implementations.
  wire_type: int

  # We cache the serialized version of the tag here so we just need to do a
  # string comparison instead of decoding the tag each time.
  encoded_tag = None

  # The semantic type of the object described by this descriptor.
  type = None

  # The type name according to the .proto domain specific language.
  proto_type_name = "string"

  # A field may be defined but not added to the container immediately. In that
  # case we wait for late binding to resolve the target and then bind the field
  # to the protobuf descriptor set only when its target is resolved.
  late_bound = False

  # The Semantic protobuf class which owns this field descriptor.
  owner = None

  # This flag indicates if the default should be set into the owner protobuf on
  # access.
  set_default_on_access = False

  def __init__(
      self,
      field_number=None,
      required=False,
      labels=None,
      set_default_on_access=None,
      **kwargs,
  ):
    super().__init__(**kwargs)
    # TODO: Without this type hint, pytype thinks that field_number
    # is always None.
    self.field_number: int = field_number
    self.required = required
    if set_default_on_access is not None:
      self.set_default_on_access = set_default_on_access

    self.labels = labels or []
    if self.field_number is None:
      raise type_info.TypeValueError("No valid field number specified.")

    self.CalculateTags()

  def Copy(self, field_number=None):
    """Returns a copy of descriptor, optionally changing the field number."""
    result = copy.copy(self)
    if field_number is not None:
      result.field_number = field_number
      result.CalculateTags()

    return result

  def CalculateTags(self):
    # In python Varint encoding is expensive so we want to move as much of the
    # hard work from the Write() methods which are called frequently to the type
    # descriptor constructor which is only called once (during protobuf
    # declaration time). Pre-calculating the tag makes for faster serialization.
    self.tag = self.field_number << 3 | self.wire_type
    self.encoded_tag = VarintEncode(self.tag)

  def IsDirty(self, unused_python_format):
    """Return and clear the dirty state of the python object."""
    return False

  def ConvertFromWireFormat(self, value, container=None):
    """Convert value from the internal type to the real type.

    When data is being parsed, it might be quicker to store it in a different
    format internally. This is because we must parse all tags, but only decode
    those fields which are being accessed.

    This function is called when we retrieve a field on access, so we only pay
    the penalty once, and cache the result.

    Internally the wire format is a tuple of:
      (encoded_tag, optional_encoded_length, encoded_data)

    All items are strings. For length encoded fields, the
    optional_encoded_length should be the encoded length of encoded_data,
    otherwise it should be the empty string. The idea is that the encoder simply
    concatenates all the wire formats together to form the final message without
    delegating to the field descriptors.

    This function is HOT.

    Args:
      value: A parameter stored in the wire format for this type.
      container: The protobuf that contains this field.

    Returns:
      The parameter encoded in the python format representation.
    """
    raise NotImplementedError

  def ConvertToWireFormat(self, value):
    """Convert the parameter into the internal storage format.

    This function is the inverse of ConvertFromWireFormat(). See the description
    above for the exact layout of the internal wire format.

    This function is HOT.

    Args:
      value: A python format representation of the value as coerced by the
        Validate() method. This is type specific, but always the same.

    Returns:
      The parameter encoded in the wire format representation.
    """
    raise NotImplementedError

  def _FormatDescriptionComment(self):
    result = "".join(["\n  // %s\n" % x for x in self.description.splitlines()])
    return result

  def _FormatDefault(self):
    return " [default = %s]" % self.GetDefault()

  def _FormatField(self):
    result = "  optional %s %s = %s%s" % (
        self.proto_type_name,
        self.name,
        self.field_number,
        self._FormatDefault(),
    )

    return result + ";\n"

  def Definition(self):
    """Return a string with the definition of this field."""
    return self._FormatDescriptionComment() + self._FormatField()

  def Format(self, value) -> Iterator[str]:
    """A Generator for display lines representing value."""
    yield str(value)

  def Validate(self, value, container=None):
    """Validate the value."""
    _ = container
    return value

  def GetDefault(self, container=None):
    _ = container
    return self.default

  def __str__(self) -> str:
    # TODO: This fails for ProtoList.
    return "<Field %s (%s) of %s: field_number: %s>" % (
        self.name,
        self.__class__.__name__,
        self.owner.__name__,
        self.field_number,
    )

  def SetOwner(self, owner):
    self.owner = owner


class ProtoString(ProtoType):
  """A string encoded in a protobuf."""

  wire_type = WIRETYPE_LENGTH_DELIMITED

  # This descriptor describes unicode strings.
  type = rdfvalue.RDFString

  def __init__(self, default="", **kwargs):
    # Strings default to "" if not specified.
    super().__init__(**kwargs)

    # Ensure the default is a unicode object.
    if default is not None:
      self.default = utils.SmartUnicode(default)

  def GetDefault(self, container=None):
    _ = container
    return self.default

  def Validate(self, value, **_) -> str:  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    """Validates a python format representation of the value."""
    if isinstance(value, rdfvalue.RDFString):
      # TODO(hanuszczak): Use `str` here.
      return str(value)

    if isinstance(value, str):
      return value

    if isinstance(value, bytes):
      return value.decode("utf-8")

    raise type_info.TypeValueError(
        "Not a valid unicode string: {!r} of type {}".format(
            value, type(value).__name__
        )
    )

  def ConvertFromWireFormat(self, value, container=None):
    """Internally strings are utf8 encoded."""
    try:
      return value[2].decode("utf8")
    except UnicodeDecodeError:
      raise rdfvalue.DecodeError("Unicode decoding error")

  def ConvertToWireFormat(self, value):
    """Internally strings are utf8 encoded."""
    value = value.encode("utf8")
    return (self.encoded_tag, VarintEncode(len(value)), value)

  def Definition(self):
    """Return a string with the definition of this field."""
    return self._FormatDescriptionComment() + self._FormatField()

  def _FormatDefault(self):
    if self.GetDefault():
      return " [default = %r]" % self.GetDefault()
    else:
      return ""

  def Format(self, value):
    yield repr(value)


class ProtoBinary(ProtoType):
  """A binary string encoded in a protobuf."""

  wire_type = WIRETYPE_LENGTH_DELIMITED

  # This descriptor describes strings.
  type = rdfvalue.RDFBytes

  proto_type_name = "bytes"

  def __init__(self, default=b"", **kwargs):
    precondition.AssertType(default, bytes)

    # Byte strings default to "" if not specified.
    super().__init__(**kwargs)

    # Ensure the default is a string object.
    if default is not None:
      self.default = default

  def Validate(self, value, **_):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    if not isinstance(value, bytes):
      raise type_info.TypeValueError("Required bytes, got %r" % value)

    return value

  def ConvertFromWireFormat(self, value, container=None):
    return value[2]

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, VarintEncode(len(value)), value)

  def Definition(self):
    """Return a string with the definition of this field."""
    return self._FormatDescriptionComment() + self._FormatField()

  def Format(self, value):
    yield repr(value)

  def _FormatDefault(self):
    if self.GetDefault():
      return " [default = %r]" % self.GetDefault()
    else:
      return ""


class ProtoUnsignedInteger(ProtoType):
  """An unsigned VarInt encoded in the protobuf."""

  wire_type = WIRETYPE_VARINT

  # This descriptor describes integers.
  type = rdfvalue.RDFInteger
  proto_type_name = "uint64"

  def __init__(self, default=0, **kwargs):
    # Integers default to 0 if not specified.
    super().__init__(default=default, **kwargs)

  def ConvertFromWireFormat(self, value, container=None):
    return VarintReader(value[2], 0)[0]

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", VarintEncode(value))

  def Validate(self, value, **_):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    try:
      return int(value)
    except ValueError:
      raise type_info.TypeValueError("Invalid value %s for Integer" % value)

  def _FormatDefault(self):
    if self.GetDefault():
      return " [default = %r]" % self.GetDefault()
    else:
      return ""


class ProtoSignedInteger(ProtoUnsignedInteger):
  """A signed VarInt encoded in the protobuf.

  Note: signed VarInts are more expensive than unsigned VarInts.
  """

  proto_type_name = "int64"

  def ConvertFromWireFormat(self, value, container=None):
    return SignedVarintReader(value[2])[0]

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", SignedVarintEncode(value))


class ProtoFixed32(ProtoUnsignedInteger):
  """A 32 bit fixed unsigned integer.

  The wire format is a 4 byte string, while the python type is an int.
  """

  _size = 4

  proto_type_name = "sfixed32"
  wire_type = WIRETYPE_FIXED32

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", struct.pack("<L", int(value)))

  def ConvertFromWireFormat(self, value, container=None):
    return struct.unpack("<L", value[2])[0]


class ProtoFixed64(ProtoFixed32):
  _size = 8

  proto_type_name = "sfixed64"
  wire_type = WIRETYPE_FIXED64

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", struct.pack("<Q", int(value)))

  def ConvertFromWireFormat(self, value, container=None):
    return struct.unpack("<Q", value[2])[0]


class ProtoFixedU32(ProtoFixed32):
  """A 32 bit fixed unsigned integer.

  The wire format is a 4 byte string, while the python type is an int.
  """

  proto_type_name = "fixed32"

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", struct.pack("<l", int(value)))

  def ConvertFromWireFormat(self, value, container=None):
    return struct.unpack("<l", value[2])[0]


class ProtoFloat(ProtoFixed32):
  """A float.

  The wire format is a 4 byte string, while the python type is a float.
  """

  proto_type_name = "float"

  def Validate(self, value, **_):
    if not rdfvalue.RDFInteger.IsNumeric(value):
      raise type_info.TypeValueError("Invalid value %s for Float" % value)

    return value

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", struct.pack("<f", float(value)))

  def ConvertFromWireFormat(self, value, container=None):
    return struct.unpack("<f", value[2])[0]


class ProtoDouble(ProtoFixed64):
  """A double.

  The wire format is a 8 byte string, while the python type is a float.
  """

  proto_type_name = "double"

  def Validate(self, value, **_):
    if not rdfvalue.RDFInteger.IsNumeric(value):
      raise type_info.TypeValueError("Invalid value %s for Integer" % value)

    return value

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", struct.pack("<d", float(value)))

  def ConvertFromWireFormat(self, value, container=None):
    return struct.unpack("<d", value[2])[0]


@functools.total_ordering
class EnumNamedValue(rdfvalue.RDFPrimitive):
  """A class that wraps enums.

  Enums are just integers, except when printed they have a name.
  """

  protobuf_type = "integer"

  def __init__(
      self, initializer=None, name=None, description=None, labels=None
  ):

    if initializer is None:
      initializer = 0

    if name is None:
      name = str(initializer)

    if labels is None:
      labels = ()

    precondition.AssertType(name, str)
    precondition.AssertOptionalType(description, str)
    precondition.AssertIterableType(labels, int)

    super().__init__((int(initializer), name, description, tuple(labels)))

  @property
  def id(self):
    return self._value[0]

  @property
  def name(self):
    return self._value[1]

  @property
  def description(self):
    return self._value[2]

  @property
  def labels(self):
    return self._value[3]

  # Required, because in Python 3 overriding `__eq__` nullifies `__hash__`.
  def __hash__(self):
    return hash(self.id)

  def __eq__(self, other):
    # `other` needs to be LHS, because future newstr does not properly return
    # NotImplemented on type mismatch.
    return other == self.id or other == self.name

  def __lt__(self, other):
    if isinstance(other, EnumNamedValue):
      return (self.id, self.name) < (other.id, other.name)
    return NotImplemented

  def __str__(self) -> str:
    return self.name

  def __repr__(self):
    return "{}(initializer={!r}, name={!r})".format(
        type(self).__name__, self.id, self.name
    )

  def __int__(self):
    return self.id

  def __index__(self):
    return self.id

  def __bool__(self):
    return bool(self.id)

  def Copy(self):
    return type(self)(self.id, self.name, self.description, self.labels)

  def SerializeToBytes(self) -> bytes:
    return str(self.id).encode("ascii")

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)

    if value:
      return cls(int(value))
    else:
      return cls(0)

  @classmethod
  def FromHumanReadable(cls, string: str):
    precondition.AssertType(string, str)

    try:
      num = int(string)
    except ValueError:
      num = 0

    return cls(num, name=string)

  @classmethod
  def FromWireFormat(cls, value):
    return cls(initializer=value)

  def SerializeToWireFormat(self):
    return self.id


class ProtoEnum(ProtoSignedInteger):
  """An enum native proto type.

  This is really encoded as an integer but only certain values are allowed.
  """

  type = EnumNamedValue

  def __init__(
      self,
      default=None,
      enum_name=None,
      enum=None,
      enum_descriptions=None,
      enum_labels=None,
      **kwargs,
  ):
    super().__init__(**kwargs)
    if enum_name is None:
      raise type_info.TypeValueError("Enum groups must be given a name.")

    self.enum_name = enum_name
    self.proto_type_name = enum_name
    if isinstance(enum, EnumContainer):
      enum = enum.enum_dict

    for v in enum.values():
      if v.__class__ is not int:
        raise type_info.TypeValueError("Enum values must be integers.")

    self.enum_container = EnumContainer(
        name=enum_name,
        descriptions=enum_descriptions,
        enum_labels=enum_labels,
        values=(enum or {}),
    )
    self.enum = self.enum_container.enum_dict
    self.reverse_enum = self.enum_container.reverse_enum

    # Ensure the default is a valid enum value.
    if default is not None:
      self.default = self.Validate(default)

  def GetDefault(self, container=None):
    _ = container
    return EnumNamedValue(
        self.default, name=self.reverse_enum.get(self.default)
    )

  def Validate(self, value, **_):
    """Check that value is a valid enum."""
    # None is a valid value - it means the field is not set.
    if value is None:
      return

    # If the value is a string we need to try to convert it to an integer.
    checked_value = value
    if isinstance(value, str):
      # NOTE: that when initializing from string, enum values are
      # case-insensitive.
      checked_value = self.enum.get(value.upper())
      if checked_value is None and value.isdigit():
        checked_value = int(value)
      if checked_value is None:
        raise type_info.TypeValueError(
            "Value %s is not a valid enum value for field %s"
            % (value, self.name)
        )

    return EnumNamedValue(checked_value, name=self.reverse_enum.get(value))

  def Definition(self):
    """Return a string with the definition of this field."""
    result = self._FormatDescriptionComment()

    result += "  enum %s {\n" % self.enum_name
    for k, v in sorted(self.reverse_enum.items()):
      result += "    %s = %s;\n" % (v, k)

    result += "  }\n"

    result += self._FormatField()
    return result

  def Format(self, value):
    yield self.reverse_enum.get(value, str(value))

  def ConvertToWireFormat(self, value):
    return (self.encoded_tag, b"", SignedVarintEncode(int(value)))

  def ConvertFromWireFormat(self, value, container=None):
    value = SignedVarintReader(value[2], 0)[0]
    return EnumNamedValue(value, name=self.reverse_enum.get(value))


class ProtoBoolean(ProtoEnum):
  """A Boolean."""

  type = bool

  def __init__(self, **kwargs):
    super().__init__(enum_name="Bool", enum={"True": 1, "False": 0}, **kwargs)

    self.proto_type_name = "bool"

  def GetDefault(self, container=None):
    """Return boolean value."""
    return bool(int(super().GetDefault(container=container)))

  def Validate(self, value, **_):
    """Check that value is a valid enum."""
    if value is None:
      return

    return bool(int(super().Validate(value)))

  def ConvertFromWireFormat(self, value, container=None):
    return bool(int(super().ConvertFromWireFormat(value, container=container)))

  def ConvertToWireFormat(self, value):
    return super().ConvertToWireFormat(bool(value))


class ProtoEmbedded(ProtoType):
  """A field may be embedded as a serialized protobuf.

  Embedding is more efficient than nesting since the emebedded protobuf does not
  need to be parsed at all, if the user does not access any elements in it.

  Embedded protobufs are simply serialized as bytes using the wire format
  WIRETYPE_LENGTH_DELIMITED. Hence the wire format is a simple python string,
  but the python format representation is an RDFProtoStruct.
  """

  wire_type = WIRETYPE_LENGTH_DELIMITED

  # When we access a nested protobuf we automatically create it and assign it to
  # the owner protobuf.
  set_default_on_access = True

  def __init__(self, nested=None, **kwargs):
    super().__init__(**kwargs)

    # Nested can refer to a target RDFProtoStruct by name.
    if isinstance(nested, str):
      self.proto_type_name = nested

      # Try to resolve the type it names
      self.type = rdfvalue.RDFValue.classes.get(nested, None)

      # We do not know about this type yet. Implement Late Binding.
      if self.type is None:
        self.late_bound = True

        # Register a late binding callback.
        rdfvalue.RegisterLateBindingCallback(nested, self.LateBind)

    # Or it can be an subclass of RDFProtoStruct.
    elif issubclass(nested, RDFProtoStruct):
      self.type = nested
      self.proto_type_name = nested.__name__

    else:
      raise type_info.TypeValueError(
          "Only RDFProtoStructs can be nested, not %s" % nested.__name__
      )

  def ConvertFromWireFormat(self, value, container=None):
    """The wire format is simply a string."""
    result = self.type()
    ReadIntoObject(value[2], 0, result)

    return result

  def ConvertToWireFormat(self, value):
    """Encode the nested protobuf into wire format."""
    output = _SerializeEntries(_GetOrderedEntries(value.GetRawData()))
    return (self.encoded_tag, VarintEncode(len(output)), output)

  def LateBind(self, target=None):
    """Late binding callback.

    This method is called on this field descriptor when the target RDFValue
    class is finally defined. It gives the field descriptor an opportunity to
    initialize after the point of definition.

    Args:
      target: The target nested class.

    Raises:
      TypeError: If the target class is not of the expected type.
    """
    if not issubclass(target, RDFProtoStruct):
      raise TypeError(
          "Field %s expects a protobuf, but target is %s" % (self, target)
      )

    self.late_bound = False

    # The target type is now resolved.
    self.type = target

    # Register us in our owner.
    self.owner.AddDescriptor(self)

  def IsDirty(self, proto):
    """Return and clear the dirty state of the python object."""
    if proto.dirty:
      return True

    for python_format, _, type_descriptor in proto.GetRawData().values():
      if python_format is not None and type_descriptor.IsDirty(python_format):
        proto.dirty = True
        return True

    return False

  def GetDefault(self, container=None):
    """When a nested proto is accessed, default to an empty one."""
    return self.type()

  def Validate(self, value, **_):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    if isinstance(value, str):
      raise type_info.TypeValueError(
          "Field %s must be of type %s" % (self.name, self.type.__name__)
      )

    # We may coerce it to the correct type.
    if value.__class__ is not self.type:
      try:
        value = self.type(value)
      except rdfvalue.InitializeError as e:
        raise type_info.TypeValueError(
            "Field %s must be of type %s" % (self.name, self.type.__name__)
        ) from e

    return value

  def Definition(self):
    """Return a string with the definition of this field."""
    return self._FormatDescriptionComment() + self._FormatField()

  def _FormatField(self):
    result = "  optional %s %s = %s" % (
        self.proto_type_name,
        self.name,
        self.field_number,
    )
    return result + ";\n"

  def Format(self, value):
    for line in value.Format():
      yield "  %s" % line


class ProtoDynamicEmbedded(ProtoType):
  """An embedded field which has a dynamic type."""

  wire_type = WIRETYPE_LENGTH_DELIMITED

  set_default_on_access = True

  proto_type_name = "bytes"

  def __init__(self, dynamic_cb=None, **kwargs):
    """Initialize the type descriptor.

    We call the dynamic_method to know which type should be used to decode the
    embedded bytestream.

    Args:
      dynamic_cb: A callback to be used to return the class to parse the
        embedded data. We pass the callback our container.
      **kwargs: Passthrough.
    """
    super().__init__(**kwargs)
    self._type = dynamic_cb

  def ConvertFromWireFormat(self, value, container=None):
    """The wire format is simply a string."""
    return serialization.FromBytes(self._type(container), value[2])

  def ConvertToWireFormat(self, value):
    """Encode the nested protobuf into wire format."""
    data = serialization.ToBytes(value)
    return (self.encoded_tag, VarintEncode(len(data)), data)

  def Validate(self, value, container=None):
    if self._type is None:
      return value

    required_type = self._type(container)
    if required_type and not isinstance(value, required_type):
      raise ValueError(
          "Expected value of type %s, but got %s"
          % (required_type, value.__class__.__name__)
      )

    return value

  def GetDefault(self, container=None):
    # _type can be None if a ProtoDynamicEmbedded field is not initialized.
    if self._type is None:
      return None

    cls = self._type(container or self.owner())
    if cls is not None:
      return cls()

  def Format(self, value):
    yield "  %r" % value

  def _FormatDefault(self):
    return ""


class ProtoDynamicAnyValueEmbedded(ProtoDynamicEmbedded):
  """An embedded dynamic field which that is stored as AnyValue struct."""

  proto_type_name = "google.protobuf.Any"

  WRAPPER_BY_TYPE = {
      "bytes": wrappers_pb2.BytesValue,
      "string": wrappers_pb2.StringValue,
      "integer": wrappers_pb2.Int64Value,
      "unsigned_integer_32": wrappers_pb2.UInt32Value,
      "unsigned_integer": wrappers_pb2.UInt64Value,
  }

  TYPE_BY_WRAPPER = {
      "BytesValue": rdfvalue.RDFBytes,
      "StringValue": rdfvalue.RDFString,
      "Int64Value": rdfvalue.RDFInteger,
      "UInt32Value": rdfvalue.RDFInteger,
      "UInt64Value": rdfvalue.RDFInteger,
  }

  def ConvertFromWireFormat(self, value, container=None):
    """The wire format is an AnyValue message."""
    result = AnyValue()
    ReadIntoObject(value[2], 0, result)
    if self._type is not None:
      converted_value = self._type(container)
    else:
      converted_value = self._TypeFromAnyValue(result)

    # If one of the protobuf library wrapper classes is used, unwrap the value.
    if result.type_url.startswith("type.googleapis.com/google.protobuf."):
      wrapper_cls = self.__class__.WRAPPER_BY_TYPE[
          converted_value.protobuf_type
      ]
      wrapper_value = wrapper_cls()
      wrapper_value.ParseFromString(result.value)
      return converted_value.FromWireFormat(wrapper_value.value)
    else:
      # TODO(user): Type stored in type_url is currently ignored when value
      # is decoded. We should use it to deserialize the value and then check
      # that value type and dynamic type are compatible.
      return converted_value.FromSerializedBytes(result.value)

  def _TypeFromAnyValue(self, anyvalue):
    type_str = anyvalue.type_url.split("/")[-1].split(".")[-1]

    if type_str in self.TYPE_BY_WRAPPER:
      return self.TYPE_BY_WRAPPER[type_str]

    try:
      return rdfvalue.RDFValue.classes[type_str]
    except KeyError:
      raise TypeError("Can't find RDFValue class for type: %s" % type_str)

  def ConvertToWireFormat(self, value):
    """Encode the nested protobuf into wire format."""

    try:
      primitive_protobuf_type = serialization.GetProtobufType(type(value))
    except ValueError:
      primitive_protobuf_type = None

    # Is it a protobuf-based value?
    if hasattr(value.__class__, "protobuf"):
      if value.__class__.protobuf:
        type_name = (
            "type.googleapis.com/grr.%s"
            % value.__class__.protobuf.__name__
        )
      else:
        type_name = value.__class__.__name__
      data = value.SerializeToBytes()
    # Is it a primitive value?
    elif primitive_protobuf_type is not None:
      wrapper_cls = self.__class__.WRAPPER_BY_TYPE[primitive_protobuf_type]
      wrapped_data = wrapper_cls()
      wrapped_data.value = serialization.ToWireFormat(value)

      type_name = (
          "type.googleapis.com/google.protobuf.%s" % wrapper_cls.__name__
      )
      data = wrapped_data.SerializeToString()
    else:
      raise ValueError(
          "Can't convert value %s to a protobuf.Any value." % value
      )

    any_value = AnyValue(type_url=type_name, value=data)
    output = _SerializeEntries(_GetOrderedEntries(any_value.GetRawData()))

    return (self.encoded_tag, VarintEncode(len(output)), output)


class ProtoAnyValue(ProtoType):
  """A raw Protocol Buffers `Any` type without dynamic typing magic."""

  wire_type = WIRETYPE_LENGTH_DELIMITED
  proto_type_name = "google.protobuf.Any"
  set_default_on_access = True

  def GetDefault(self, container=None):
    del container  # Unused.
    return AnyValue()

  def Validate(self, value, container=None):
    del container  # Unused.
    precondition.AssertType(value, AnyValue)

    return value

  def ConvertFromWireFormat(self, value, container=None):
    del container  # Unused.

    return AnyValue.FromSerializedBytes(value[2])

  def ConvertToWireFormat(self, value):
    precondition.AssertType(value, AnyValue)

    data = value.SerializeToBytes()
    return (self.encoded_tag, VarintEncode(len(data)), data)


class RepeatedFieldHelper(abc.Sequence):
  """A helper for the RDFProto to handle repeated fields.

  This helper is intended to only be constructed from the RDFProto class.
  """

  def __init__(self, wrapped_list=None, type_descriptor=None, container=None):
    """Constructor.

    Args:
      wrapped_list: The list within the protobuf which we wrap.
      type_descriptor: A type descriptor describing the type of the list
        elements.
      container: The protobuf which contains this repeated field.

    Raises:
      AttributeError: If parameters are not valid.
    """
    self.dirty = False

    if wrapped_list is None:
      self.wrapped_list = []

    # TODO(user): type checker doesn't respect the check below
    # and doesn't use it to infer the type for wrapped_list.
    elif wrapped_list.__class__ is RepeatedFieldHelper:
      self.wrapped_list = cast(RepeatedFieldHelper, wrapped_list).wrapped_list

    else:
      self.wrapped_list = wrapped_list

    if type_descriptor is None:
      raise AttributeError("type_descriptor not specified.")

    self.type_descriptor = type_descriptor
    self.container = container

  def IsDirty(self):
    """Is this repeated item dirty?

    This is used to invalidate any caches that our owners have of us.

    Returns:
      True if this object is dirty.
    """
    if self.dirty:
      return True

    # If any of the items is dirty we are also dirty.
    for item in self.wrapped_list:
      if self.type_descriptor.IsDirty(item[0]):
        self.dirty = True
        return True

    return False

  def Copy(self):
    return RepeatedFieldHelper(
        wrapped_list=self.wrapped_list[:], type_descriptor=self.type_descriptor
    )

  def Append(self, rdf_value=utils.NotAValue, wire_format=None, **kwargs):
    """Append the value to our internal list."""
    if rdf_value is utils.NotAValue:
      if wire_format is None:
        rdf_value = self.type_descriptor.type(**kwargs)
        self.dirty = True
      else:
        rdf_value = None
    else:
      # Coerce the value to the required type.
      try:
        rdf_value = self.type_descriptor.Validate(rdf_value, **kwargs)
      except (TypeError, ValueError) as e:
        raise type_info.TypeValueError(
            "Assignment value must be %s, but %s can not be coerced. Error: %s"
            % (self.type_descriptor.proto_type_name, type(rdf_value), e)
        )

    self.wrapped_list.append((rdf_value, wire_format))

    return rdf_value

  def Pop(self, item):
    result = self[item]
    self.wrapped_list.pop(item)
    return result

  def Extend(self, iterable):
    for i in iterable:
      self.Append(rdf_value=i)

  append = utils.Proxy("Append")
  remove = utils.Proxy("Remove")
  extend = utils.Proxy("Extend")

  def __getitem__(self, item):
    # Ensure we handle slices as well.
    if item.__class__ is slice:
      result = []
      for i in range(*item.indices(len(self))):
        result.append(self.wrapped_list[i])

      return self.__class__(
          wrapped_list=result, type_descriptor=self.type_descriptor
      )

    python_format, wire_format = self.wrapped_list[item]
    if python_format is None:
      python_format = self.type_descriptor.ConvertFromWireFormat(
          wire_format, container=self.container
      )

      self.wrapped_list[item] = (python_format, wire_format)

    return python_format

  def __len__(self):
    return len(self.wrapped_list)

  def __ne__(self, other):
    return not self == other  # pylint: disable=g-comparison-negation

  def __eq__(self, other):
    if not isinstance(other, Sequence) or isinstance(other, (ByteString, str)):
      return NotImplemented
    if len(self) != len(other):
      return False
    for x, y in zip(self, other):
      if x != y:
        return False
    return True

  def __str__(self) -> str:
    result = ["'%s': [" % self.type_descriptor.name]
    for element in self:
      for line in self.type_descriptor.Format(element):
        result.append(" %s" % line)

    result.append("]")

    return "\n".join(result)

  def __repr__(self):
    # Skip self.container and self.type_descriptor to avoid cyclical output.
    return "<{} {!r}>".format(type(self).__name__, self.wrapped_list)

  def Validate(self):
    for x in self:
      if hasattr(x, "Validate"):
        x.Validate()


class ProtoList(ProtoType):
  """A repeated type."""

  #         ,     \    /      ,
  #        / \    )\__/(     / \
  #       /   \  (_\  /_)   /   \
  #  ____/_____\__\@  @/___/_____\____
  # |             |\../|              |
  # |              \VV/               |
  # |    WARNING: Here be dragons!    |
  # | When accessing a ProtoList that |
  # | is a field of another RDFValue, |
  # | its unset value is  replaced    |
  # | with its default value: [].     |
  # | Since [] is not None, accessing |
  # | a ProtoList changes its parents |
  # | equality.                       |
  # |_________________________________|
  #  |    /\ /      \\       \ /\    |
  #  |  /   V        ))       V   \  |
  #  |/     `       //        '     \|
  #  `              V                '
  # TODO
  set_default_on_access = True

  def __init__(self, delegate, labels=None, **kwargs):
    self.delegate = delegate
    if not isinstance(delegate, ProtoType):
      raise AttributeError(
          "Delegate class must derive from ProtoType, not %s"
          % delegate.__class__.__name__
      )

    # If our delegate is late bound we must also be late bound. This means that
    # the repeated field is not registered in the owner protobuf just
    # yet. However, we do not actually need to register a late binding callback
    # ourselves, since the delegate field descriptor already did this. We simply
    # wait until the delegate calls our AddDescriptor() method and then we call
    # our own owner's AddDescriptor() method to ensure we re-register.
    self.late_bound = delegate.late_bound

    self.wire_type = delegate.wire_type

    super().__init__(
        name=delegate.name,
        description=delegate.description,
        field_number=delegate.field_number,
        friendly_name=delegate.friendly_name,
        labels=labels,
    )

  def IsDirty(self, value):
    return value.IsDirty()

  def GetDefault(self, container=None):
    # By default an empty RepeatedFieldHelper.
    return RepeatedFieldHelper(
        type_descriptor=self.delegate, container=container
    )

  def Validate(self, value, **_):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    """Check that value is a list of the required type."""
    # Assigning from same kind can allow us to skip verification since all
    # elements in a RepeatedFieldHelper already are coerced to the delegate
    # type. In that case we just make a copy. This only works when the value
    # wraps the same type as us.
    if (
        value.__class__ is RepeatedFieldHelper
        and value.type_descriptor is self.delegate
    ):
      result = value.Copy()

    # Make sure the base class finds the value valid.
    else:
      # The value may be a generator here, so we just iterate over it.
      result = RepeatedFieldHelper(type_descriptor=self.delegate)
      result.Extend(value)

    return result

  def ConvertFromWireFormat(self, value, container=None):
    result = RepeatedFieldHelper(type_descriptor=self.delegate)
    for wire_format in SplitBuffer(value[2]):
      result.wrapped_list.append((None, wire_format))

    return result

  def ConvertToWireFormat(self, value):
    """Convert to the wire format.

    Args:
      value: is of type RepeatedFieldHelper.

    Returns:
      A wire format representation of the value.
    """
    output = _SerializeEntries(
        (python_format, wire_format, value.type_descriptor)
        for (python_format, wire_format) in value.wrapped_list
    )
    return b"", b"", output

  def Format(self, value):
    yield "["

    for element in value:
      for line in self.delegate.Format(element):
        yield " %s" % line

    yield "]"

  def _FormatField(self):
    result = "  repeated %s %s = %s" % (
        self.delegate.proto_type_name,
        self.name,
        self.field_number,
    )

    return result + ";\n"

  def SetOwner(self, owner):
    self.owner = owner
    # We are the owner for the delegate field descriptor.
    self.delegate.SetOwner(self)

  def AddDescriptor(self, field_desc):
    """This method will be called by our delegate during late binding."""
    # Just relay it up to our owner.
    self.late_bound = False
    self.delegate = field_desc
    self.wire_type = self.delegate.wire_type
    self.owner.AddDescriptor(self)


class ProtoRDFValue(ProtoType):
  """Serialize arbitrary rdfvalue members.

  RDFValue members can be serialized in a number of different ways according to
  their preferred protobuf_type member. We map the descriptions in
  protobuf_type into a suitable protobuf serialization for optimal
  serialization. We therefore use a delegate type descriptor to best convert
  from the RDFValue to the wire type. For example, an RDFDatetime is best
  represented as an integer (number of microseconds since the epoch). Hence
  RDFDatetime.SerializeToWireFormat() will return an integer, and the delegate
  will be ProtoUnsignedInteger().

  To convert from the RDFValue python type to the delegate's wire type we
  therefore need to make two conversions:

  1) Our python format is the RDFValue -> intermediate data store format using
  RDFValue.SerializeToWireFormat(). This will produce a python object which is
  the correct python format for the delegate primitive type descriptor.

  2) Use the delegate to obtain the wire format of its own python type
  (i.e. self.delegate.ConvertToWireFormat())

  NOTE: The default value for an RDFValue is None. It is impossible for us to
  know how to instantiate a valid default value without being told by the
  user. This is unlike the default value for strings or ints which are "" and 0
  respectively.
  """

  # We delegate encoding/decoding to a primitive field descriptor based on the
  # semantic type's protobuf_type attribute.
  primitive_desc = None

  # We store our args here so we can use the same args to initialize the
  # delegate descriptor.
  _kwargs = None

  type = None
  wire_type = WIRETYPE_LENGTH_DELIMITED

  _PROTO_DATA_STORE_LOOKUP = dict(
      bytes=ProtoBinary,
      unsigned_integer=ProtoUnsignedInteger,
      unsigned_integer_32=ProtoUnsignedInteger,
      integer=ProtoUnsignedInteger,
      signed_integer=ProtoSignedInteger,
      string=ProtoString,
  )

  def __init__(self, rdf_type=None, default=None, **kwargs):
    super().__init__(**kwargs)
    self._kwargs = kwargs

    if default is not None:
      self.default = default

    if isinstance(rdf_type, str):
      self.original_proto_type_name = self.proto_type_name = rdf_type

      # Try to resolve the type it names
      self.type = rdfvalue.RDFValue.classes.get(rdf_type, None)

      # We do not know about this type yet. Implement Late Binding.
      if self.type is None:
        self.late_bound = True

        # Register a late binding callback.
        rdfvalue.RegisterLateBindingCallback(rdf_type, self.LateBind)

      else:
        # The semantic type was found successfully.
        self._GetPrimitiveEncoder()

    # Or it can be an subclass of RDFValue.
    elif issubclass(rdf_type, rdfvalue.RDFValue):  # pytype: disable=wrong-arg-types
      self.type = rdf_type
      self.original_proto_type_name = self.proto_type_name = rdf_type.__name__
      self._GetPrimitiveEncoder()

    else:
      type_info.TypeValueError("An rdf_type must be specified.")

  def LateBind(self, target=None):
    """Bind the field descriptor to the owner once the target is defined."""
    self.type = target
    self._GetPrimitiveEncoder()

    # Now re-add the descriptor to the owner protobuf.
    self.late_bound = False
    self.owner.AddDescriptor(self)

  def _GetPrimitiveEncoder(self):
    """Finds the primitive encoder according to the type's protobuf_type."""
    # Decide what should the primitive type be for packing the target rdfvalue
    # into the protobuf and create a delegate descriptor to control that.
    primitive_cls = self._PROTO_DATA_STORE_LOOKUP[
        serialization.GetProtobufType(self.type)
    ]
    self.primitive_desc = primitive_cls(**self._kwargs)

    # Our wiretype is the same as the delegate's.
    self.wire_type = self.primitive_desc.wire_type
    self.proto_type_name = self.primitive_desc.proto_type_name

    # Recalculate our tags.
    self.CalculateTags()

  def GetDefault(self, container=None):
    _ = container
    if self.default is None:
      return None

    if self.default.__class__ is not self.type:
      self.default = self.Validate(self.default)
    return self.default

  def IsDirty(self, python_format):
    """Return the dirty state of the python object."""
    return python_format.dirty

  def Definition(self):
    return (
        "\n  // Semantic Type: %s" % self.type.__name__
    ) + self.primitive_desc.Definition()

  def Validate(self, value, **_):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    # Try to coerce into the correct type:
    if value.__class__ is not self.type:
      try:
        value = self.type(value)
      except (rdfvalue.DecodeError, TypeError) as e:
        raise type_info.TypeValueError(e)

    return value

  def ConvertFromWireFormat(self, value, container=None):
    # Wire format should be compatible with the protobuf_type for the
    # rdfvalue. We use the delegate primitive descriptor to perform the
    # conversion.
    value = self.primitive_desc.ConvertFromWireFormat(
        value, container=container
    )

    result = self.type(value)

    return result

  def ConvertToWireFormat(self, value):
    return self.primitive_desc.ConvertToWireFormat(
        value.SerializeToWireFormat()
    )

  def Copy(self, field_number=None):
    """Returns descriptor copy, optionally changing field number."""
    new_args = self._kwargs.copy()
    if field_number is not None:
      new_args["field_number"] = field_number

    return ProtoRDFValue(
        rdf_type=self.original_proto_type_name,
        default=getattr(self, "default", None),
        **new_args,
    )

  def _FormatField(self):
    result = "  optional %s %s = %s" % (
        self.proto_type_name,
        self.name,
        self.field_number,
    )
    return result + ";\n"

  def Format(self, value):
    yield "%s:" % self.type.__name__
    for line in str(value).splitlines():
      yield "  %s" % line

  def __str__(self):
    return "<Field %s (Sem Type: %s) of %s: field_number: %s>" % (
        self.name,
        self.proto_type_name,
        self.owner.__name__,
        self.field_number,
    )


class RDFStructMetaclass(rdfvalue.RDFValueMetaclass):
  """A metaclass which registers new RDFProtoStruct instances."""

  _HAS_DYNAMIC_ATTRIBUTES = True  # help out pytype

  def __init__(untyped_cls, name, bases, env_dict):  # pylint: disable=no-self-argument
    super().__init__(name, bases, env_dict)

    # TODO(user):pytype: find a more elegant solution (if possible).
    # cast() doesn't accept forward references and argument annotations
    # like Type["RDFStruct"] are not accepted by the type checker. The
    # biggest caveat here is that RDFStruct is defined *with the help*
    # of RDFStructMetaclass, so its name is not defined at the time
    # this code is evaluated.
    cls: type["RDFStruct"] = untyped_cls
    cls.type_infos = type_info.TypeDescriptorSet()

    # Keep track of the late bound fields.
    cls.late_bound_type_infos = {}

    cls.type_infos_by_field_number = {}
    cls.type_infos_by_encoded_tag = {}

    # Build the class by parsing an existing protobuf class.
    if cls.protobuf is not None:
      registry.RegisterRDFProtoStruct(cls.__name__, cls)
      rdf_proto2.DefineFromWireFormat(cls, cls.protobuf)

    # Pre-populate the class using the type_infos class member.
    if cls.type_description is not None:
      for field_desc in cls.type_description:
        cls.AddDescriptor(field_desc)

    cls._class_attributes = set(dir(cls))  # pylint: disable=protected-access


T = TypeVar("T")


class RDFStruct(rdfvalue.RDFValue, metaclass=RDFStructMetaclass):  # pylint: disable=invalid-metaclass
  """An RDFValue object which contains fields like a struct.

  Struct members contain values such as integers, strings etc. These are stored
  in an internal data structure.

  A value can be in two states, the wire format is a serialized format closely
  resembling the state it appears on the wire. The Decoded format is the
  representation closely representing an internal python type. The idea is that
  converting from a serialized wire encoding to the wire format is as cheap as
  possible. Similarly converting from a python object to the python
  representation is also very cheap.

  Lazy evaluation occurs when we need to obtain the python representation of a
  decoded field. This allows us to skip the evaluation of complex data.

  For example, suppose we have a protobuf with several "string" fields
  (i.e. unicode objects). The wire format for a "string" field is a UTF8 encoded
  binary string, but the python object is a unicode object.

  Normally when parsing the protobuf we can extract the wire format
  representation very cheaply, but conversion to a unicode object is quite
  expensive. If the user never access the specific field, we can keep the
  internal representation in wire format and not convert it to a unicode object.
  """

  # This can be populated with a type_info.TypeDescriptorSet() object to
  # initialize the class.
  type_description = None

  # This class can be defined using the protobuf definition language (e.g. a
  # .proto file). If defined here, we parse the .proto file for the message with
  # the exact same class name and add the field descriptions from it.
  definition = None

  # This class can be defined in terms of an existing annotated regular
  # protobuf. See RDFProtoStruct.DefineFromWireFormat().
  protobuf = None

  # This is where the type infos are constructed.
  type_infos: type_info.TypeDescriptorSet = None

  # Mark as dirty each time we modify this object.
  dirty = False

  # Stores the raw data here.
  _data = None

  def __init__(self, initializer=None, **kwargs):
    super().__init__()

    # Maintain the order so that parsing and serializing a proto does not change
    # the serialized form.
    self._data = {}
    self._prev_hash = None

    for arg, value in kwargs.items():
      if not hasattr(self.__class__, arg):
        if arg in self.late_bound_type_infos:
          raise AttributeError(
              "Field %s refers to an as yet undefined Semantic Type."
              % self.late_bound_type_infos[arg]
          )

        raise AttributeError(
            "Proto %s has no field %s" % (self.__class__.__name__, arg)
        )

      # Call setattr to allow the class to define @property pseudo fields which
      # can also be initialized.
      setattr(self, arg, value)

    if initializer is None:
      return

    elif initializer.__class__ is self.__class__:
      self.CopyConstructor(initializer)

    else:
      raise ValueError(
          "%s can not be initialized from %s"
          % (self.__class__.__name__, type(initializer))
      )

  def CopyConstructor(self, other):
    """Efficiently copy from other into this object.

    Basic RDFStruct objects are fully represented by their internal raw data. We
    can easily copy it by simply making a direct copy of the wire format of each
    field, and not preserving the python format. On access in the copy the field
    will be re-parsed into a python object.

    Args:
      other: An instance of the same type of this class.
    """
    self._data = {}
    for name, (obj, serialized, t_info) in other.GetRawData().items():
      if serialized is None:
        serialized = t_info.ConvertToWireFormat(obj)

      self._data[name] = (None, serialized, t_info)

  def Clear(self):
    """Clear all the fields."""
    self._data = {}

  def HasField(self, field_name):
    """Checks if the field exists."""
    return field_name in self._data

  def _CopyRawData(self):
    new_raw_data = {}

    # We need to copy all entries in _data. Those entries are tuples of
    # - an object (if it has already been deserialized)
    # - the serialized object (if it has been serialized)
    # - the type_info.
    # To copy this, it's easiest to just copy the serialized object if it
    # exists. We have to make sure though that the object is not a protobuf.
    # If it is, someone else might have changed the subobject and the
    # serialization is not accurate anymore. This is indicated by the dirty
    # flag. Type_infos can be just copied by reference.
    for name, (obj, serialized, t_info) in self._data.items():
      if serialized is None:
        obj = copy.copy(obj)
      else:
        try:
          if t_info.IsDirty(obj):
            obj, serialized = copy.copy(obj), None
          else:
            obj = None
        except AttributeError:
          obj = None

      new_raw_data[name] = (obj, serialized, t_info)
    return new_raw_data

  def Copy(self: T) -> T:
    """Make an efficient copy of this protobuf."""
    result = self.__class__()
    result.SetRawData(self._CopyRawData())
    return result

  def __deepcopy__(self, memo):
    result = self.__class__()
    result.SetRawData(copy.deepcopy(self._data, memo))

    return result

  def GetRawData(self):
    """Retrieves the raw python representation of the object.

    External users should not make use of the internal raw data structures.

    Returns:
      the raw python object representation (a dict).
    """
    return self._data

  def ListSetFields(self):
    """Iterates over the fields which are actually set.

    Yields:
      a tuple of (type_descriptor, value) for each field which is set.
    """
    for type_descriptor in self.type_infos:
      if type_descriptor.name in self._data:
        yield type_descriptor, self.Get(type_descriptor.name)

  def SetRawData(self, data):
    self._data = data
    self.dirty = True

  def SerializeToBytes(self):
    return _SerializeEntries(_GetOrderedEntries(self._data))

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    instance = cls()

    try:
      ReadIntoObject(value, 0, instance)
    except ValueError:
      logging.error(
          "Error in ReadIntoObject. %d bytes, extract: %r",
          len(value),
          value[:1000],
      )
      raise

    instance.dirty = True
    return instance

  @classmethod
  def FromWireFormat(cls, value):
    precondition.AssertType(value, bytes)
    return cls.FromSerializedBytes(value)

  __hash__ = None

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return NotImplemented

    # Compare all fields set in `self` or `other`, skipping only fields that
    # are unset in both `self` and `other`. We deliberately check fields that
    # are set in one class and not in the other. This allows RDFStructs to be
    # considered equal, when one has a field set to its default value and the
    # other has the field unset (where .Get implicitly returns the default
    # value then).
    for field in set(self.GetRawData()) | set(other.GetRawData()):

      # Deserializing Protobuf of newer versions can put `_unknown_field_{n}` in
      # _data. This is a special field without type descriptor, thus raising an
      # error in Get(). To follow Protobuf semantics, ignore fields that do not
      # have a type descriptor.
      # pylint: disable=unsupported-membership-test
      if field not in self.type_infos:
        continue

      self_val = self.Get(field, allow_set_default=False)
      other_val = other.Get(field, allow_set_default=False)

      if self_val != other_val:
        return False

    return True

  def __ne__(self, other):
    return not self == other  # pylint: disable=g-comparison-negation

  def Format(self):
    """Format a message in a human readable way."""
    yield "message %s {" % self.__class__.__name__

    for k, (python_format, wire_format, type_descriptor) in sorted(
        self.GetRawData().items()
    ):
      if python_format is None:
        python_format = type_descriptor.ConvertFromWireFormat(
            wire_format, container=self
        )

      # Skip printing of unknown fields.
      if isinstance(k, str):
        prefix = "{} :".format(k)
        for line in type_descriptor.Format(python_format):
          yield " %s %s" % (prefix, line)
          prefix = ""

    yield "}"

  def __str__(self) -> str:
    return "\n".join(self.Format())

  def __dir__(self):
    """Add the virtualized fields to the console's tab completion."""
    return dir(super()) + [x.name for x in self.type_infos]  # pylint: disable=not-an-iterable

  def _Set(self, value, type_descriptor):
    """Validate the value and set the attribute with it."""
    attr = type_descriptor.name
    prev_value = self.Get(attr, allow_set_default=False)

    # A value of None means we clear the field.
    if value is None:
      self._data.pop(attr, None)
    else:
      # Validate the value and obtain the python format representation.
      value = type_descriptor.Validate(value, container=self)

      # Store the lazy value object.
      self._data[attr] = (value, None, type_descriptor)

    # Make sure to invalidate our parent's cache if needed.
    self.dirty = True

    if self._prev_hash is not None and prev_value != self.Get(
        attr, allow_set_default=False
    ):
      try:
        hash(self)  # Recompute hash to raise if hash changed due to mutation.
      except AssertionError as ex:
        raise AssertionError(
            "Cannot set {}.{} to {} with previous value {}! hash() has "
            "changed after it has been used! Usage of RDFStructs as members of "
            "sets or keys of dicts is discouraged. If used anyway, mutating is "
            "prohibited, because it causes the hash to change. Be aware that "
            "accessing unset fields can trigger a mutation.".format(
                type(self).__name__, attr, value, prev_value
            )
        ) from ex
    return value

  def Set(self, attr, value):
    """Sets the attribute in to the value."""
    type_descriptor = self._GetTypeDescriptor(attr)
    return self._Set(value, type_descriptor)

  def _GetTypeDescriptor(self, attr):
    type_descriptor = self.type_infos.get(attr)

    if type_descriptor is None:
      raise AttributeError(
          "'%s' object has no attribute '%s'" % (self.__class__.__name__, attr)
      )

    return type_descriptor

  def Get(self, attr, allow_set_default=True):
    """Retrieve the attribute specified.

    Arguments:
      attr: String of the attribute's name
      allow_set_default: If True and attr is currently unset, permanently modify
        the instance by setting attr to the default value.

    Returns:
      The attribute's value, or the attribute's type's default value, if unset.
    """
    entry = self._data.get(attr)
    # We don't have this field, try the defaults.
    if entry is None:
      type_descriptor = self._GetTypeDescriptor(attr)
      default = type_descriptor.GetDefault(container=self)

      if default is None:
        return

      # Assign the default value now.
      if allow_set_default and type_descriptor.set_default_on_access:
        default = self.Set(attr, default)

      return default

    python_format, wire_format, type_descriptor = entry

    # Decode on demand and cache for next time.
    if python_format is None:
      python_format = type_descriptor.ConvertFromWireFormat(
          wire_format, container=self
      )

      self._data[attr] = (python_format, wire_format, type_descriptor)

    return python_format

  def ClearFieldsWithLabel(self, label, exceptions=None):
    exceptions = exceptions or []
    for desc, value in self.ListSetFields():
      if desc.name not in exceptions and label in desc.labels:
        self.Set(desc.name, None)
      else:
        if hasattr(value, "ClearFieldsWithLabel"):
          value.ClearFieldsWithLabel(label, exceptions=exceptions)

  @classmethod
  def AddDescriptor(cls, field_desc):
    if not isinstance(field_desc, ProtoType):
      raise type_info.TypeValueError(
          "%s field '%s' should be of type ProtoType"
          % (cls.__name__, field_desc.name)
      )

    cls.type_infos_by_field_number[field_desc.field_number] = field_desc
    cls.type_infos.Append(field_desc)


class EnumContainer:
  """A data class to hold enum objects."""

  def __init__(
      self, name=None, descriptions=None, enum_labels=None, values=None
  ):
    descriptions = descriptions or {}
    enum_labels = enum_labels or {}
    values = values or {}

    self.enum_dict = {}
    self.reverse_enum = {}
    self.name = name

    for k, v in values.items():
      v = EnumNamedValue(
          v,
          name=k,
          description=descriptions.get(k, None),
          labels=enum_labels.get(k, None),
      )
      self.enum_dict[k] = v
      self.reverse_enum[v] = k
      setattr(self, k, v)

  def FromInt(self, v):
    return getattr(self, self.reverse_enum[v])

  def FromString(self, v: str) -> Optional[EnumNamedValue]:
    if not v:
      return None
    if v not in self.enum_dict:
      raise ValueError(f"Invalid value {v} for {self.name}.")
    return getattr(self, v)


class RDFProtoStruct(RDFStruct):
  """An RDFStruct which uses protobufs for serialization.

  This implementation is faster than the standard protobuf library.
  """

  # TODO(user): if a semantic proto defines a field with the same name as
  # these class variables under some circumstances the proto default value will
  # be set incorrectly.  Figure out a way to make this safe.

  shortest_encoded_tag = 0
  longest_encoded_tag = 0

  # If set to a standard proto2 generated class, we introspect it and extract
  # type descriptors from it. This allows this implementation to use an
  # annotated .proto file to define semantic types.
  protobuf = None

  # This mapping is used to provide concrete implementations for semantic types
  # annotated in the .proto file. This is a dict with keys being the semantic
  # names, and values being the concrete implementations for these types.

  # All RDFValue classes used by this proto have to be specified here.

  # Recorded dependencies - used by update_rdf_deps script to update
  # dependencies setting.
  recorded_rdf_deps = None

  def __init__(self, initializer=None, **kwargs):
    super().__init__(initializer, **kwargs)

  def AsPrimitiveProto(self):
    """Return an old style protocol buffer object."""
    if self.protobuf:
      result = self.protobuf()
      result.ParseFromString(self.SerializeToBytes())
      return result

  def AsDict(self):
    result = {}
    for desc in self.type_infos:
      if self.HasField(desc.name):
        result[desc.name] = self.Get(desc.name)

    return result

  def FromDict(self, dictionary):
    """Initializes itself from a given dictionary."""
    dynamic_fields = []

    for key, value in dictionary.items():
      field_type_info = self.type_infos.get(key)
      if isinstance(field_type_info, ProtoEmbedded):
        nested_value = field_type_info.GetDefault(container=self)
        nested_value.FromDict(value)
        self.Set(key, nested_value)
      elif isinstance(field_type_info, ProtoList):
        if isinstance(field_type_info.delegate, ProtoEmbedded):
          nested_values = []
          for v in value:
            nested_value = field_type_info.delegate.GetDefault(container=self)
            nested_value.FromDict(v)
            nested_values.append(nested_value)

          self.Set(key, nested_values)
        else:
          self.Set(key, value)
      elif isinstance(field_type_info, ProtoDynamicEmbedded):
        dynamic_fields.append(field_type_info)
      elif field_type_info.proto_type_name == "bytes":
        self.Set(key, base64.decodebytes((value or "").encode("ascii")))
      else:
        self.Set(key, value)

    # Process dynamic fields after all other fields, because most probably
    # their class is determined by one of the previously set fields.
    for dynamic_field in dynamic_fields:
      nested_value = dynamic_field.GetDefault(container=self)
      if nested_value is None:
        raise RuntimeError(
            "Can't initialize dynamic field %s, probably some "
            "necessary fields weren't supplied."
            % dynamic_field.name
        )
      nested_value.FromDict(dictionary[dynamic_field.name])
      self.Set(dynamic_field.name, nested_value)

  def ToPrimitiveDict(self, stringify_leaf_fields=False):
    return self._ToPrimitive(self.AsDict(), stringify_leaf_fields)

  def _ToPrimitive(self, value, stringify_leaf_fields):
    if isinstance(value, RepeatedFieldHelper):
      return list(self._ToPrimitive(v, stringify_leaf_fields) for v in value)
    # Hack to avoid dependency loop. Safe because if value is a protodict.Dict,
    # then protodict has already been loaded.
    # TODO(user): remove this hack
    elif "Dict" in rdfvalue.RDFValue.classes and isinstance(
        value, rdfvalue.RDFValue.classes["Dict"]
    ):
      primitive_dict = {}
      # TODO(user):pytype: get rid of a dependency loop described above and
      # do a proper type check.
      for k, v in value.ToDict().items():  # pytype: disable=attribute-error
        primitive_dict[k] = self._ToPrimitive(v, stringify_leaf_fields)
      return primitive_dict
    elif isinstance(value, dict):
      primitive_dict = {}
      for k, v in value.items():
        primitive_dict[k] = self._ToPrimitive(v, stringify_leaf_fields)
      return primitive_dict
    elif isinstance(value, RDFProtoStruct):
      return self._ToPrimitive(value.AsDict(), stringify_leaf_fields)
    elif isinstance(value, (EnumNamedValue)):
      return str(value)
    elif isinstance(value, rdfvalue.RDFBytes):
      return base64.b64encode(value.SerializeToBytes()).decode("ascii")
    else:
      if stringify_leaf_fields:
        if isinstance(value, bytes):
          return base64.b64encode(value).decode("ascii")
        else:
          return str(value)
      else:
        return value

  def __bool__(self):
    return bool(self._data)

  @classmethod
  def EmitProto(cls):
    """Emits .proto file definitions."""
    result = "message %s {\n" % cls.__name__
    for _, desc in sorted(cls.type_infos_by_field_number.items()):
      result += desc.Definition()

    result += "}\n"
    return result

  def Validate(self):
    """Validates the semantic protobuf for internal consistency.

    Derived classes can override this method to ensure the proto is sane
    (e.g. required fields, or any arbitrary condition). This method is called
    prior to serialization. Note that it is not necessary to validate fields
    against their semantic types - it is impossible to set fields which are
    invalid. This function is more intended to validate the entire protobuf for
    internal consistency.

    Raises:
      type_info.TypeValueError if the proto is invalid.
    """

  @classmethod
  def FromTextFormat(cls, text):
    """Parse this object from a text representation."""
    if cls.protobuf is None:
      raise ValueError("protobuf must be set on cls.")
    tmp = cls.protobuf()  # pylint: disable=not-callable
    text_format.Merge(text, tmp)

    return cls.FromSerializedBytes(tmp.SerializeToString())

  @classmethod
  def AddDescriptor(cls, field_desc):
    """Register this descriptor with the Proto Struct."""
    if not isinstance(field_desc, ProtoType):
      raise type_info.TypeValueError(
          "%s field '%s' should be of type ProtoType"
          % (cls.__name__, field_desc.name)
      )

    # Ensure the field descriptor knows the class that owns it.
    field_desc.SetOwner(cls)

    # If the field is late bound we do not really add it to the descriptor set
    # yet. We must wait for the LateBindingPlaceHolder() to add it later.
    if field_desc.late_bound:
      # Keep track of unbound fields.
      cls.late_bound_type_infos[field_desc.name] = field_desc
      return

    # Ensure this field number is unique:
    if field_desc.field_number in cls.type_infos_by_field_number:
      raise type_info.TypeValueError(
          "Field number %s for field %s is not unique in %s"
          % (field_desc.field_number, field_desc.name, cls.__name__)
      )

    # We store an index of the type info by tag values to speed up parsing.
    cls.type_infos_by_field_number[field_desc.field_number] = field_desc
    cls.type_infos_by_encoded_tag[field_desc.encoded_tag] = field_desc

    cls.type_infos.Append(field_desc)
    cls.late_bound_type_infos.pop(field_desc.name, None)

    # Add direct accessors only if the class does not already have them.
    if not hasattr(cls, field_desc.name):
      # This lambda is a class method so pylint: disable=protected-access
      # This is much faster than __setattr__/__getattr__
      setattr(
          cls,
          field_desc.name,
          property(
              lambda self: self.Get(field_desc.name),
              lambda self, x: self._Set(x, field_desc),
              None,
              field_desc.description,
          ),
      )

  def UnionCast(self):
    union_field = getattr(self, self.union_field)
    cast_field_name = str(union_field).lower()

    set_fields = set(
        type_descriptor.name for type_descriptor, _ in self.ListSetFields()
    )

    union_cases = [
        case.lower()
        for case in self.type_infos[self.union_field].enum_container.enum_dict
    ]

    mismatched_union_cases = set_fields.intersection(union_cases).difference(
        [cast_field_name]
    )

    if mismatched_union_cases:
      raise ValueError(
          "Inconsistent union proto data. Expected only %r "
          "to be set, %r are also set."
          % (cast_field_name, list(mismatched_union_cases))
      )

    try:
      return getattr(self, cast_field_name)
    except AttributeError:
      raise AttributeError("union_field not initialized.")


class SemanticDescriptor(RDFProtoStruct):
  """A semantic protobuf describing the .proto extension."""

  protobuf = semantic_pb2.SemanticDescriptor


_V = TypeVar("_V", bound=rdfvalue.RDFValue)


class AnyValue(RDFProtoStruct):
  """Protobuf with arbitrary serialized proto and its type."""

  protobuf = any_pb2.Any
  allow_custom_class_name = True

  @classmethod
  def FromProto2(cls, proto2_any: any_pb2.Any) -> "AnyValue":
    """Converts a proto2 `Any` message to the RDF wrapper."""
    result = cls()
    result.type_url = proto2_any.type_url
    result.value = proto2_any.value
    return result

  @classmethod
  def PackProto2(cls, message: proto2_message.Message) -> "AnyValue":
    """Packs the given proto2 message into the `Any` RDF wrapper.

    Args:
      message: A message to pack.

    Returns:
      An instance of RD wrapper for `Any` with packaged message.
    """
    proto2_any = any_pb2.Any()
    proto2_any.Pack(message)

    return cls.FromProto2(proto2_any)

  @classmethod
  def Pack(cls, value: RDFProtoStruct) -> "AnyValue":
    """Packs given RDF value into the `Any` RDF wrapper.

    Args:
      value: An RDF value to pack.

    Returns:
      An instance of RDF wrapper for `Any` with packed message.
    """
    result = cls()
    result.type_url = TypeURL(type(value))
    result.value = value.SerializeToBytes()
    return result

  def Unpack(self, cls: type[_V]) -> _V:
    """Unpacks `Any` into an instance of the specified RDF class.

    Args:
      cls: A class into instance of which the value should be unpacked into.

    Returns:
      An instance of the specified class.
    """
    # Because messages can get renamed, unpacking into an unexpected type is not
    # necessarily an error, but it still worth to log it as it might be helpful
    # in identifying issues.
    cls_type_url = TypeURL(cls)
    if cls_type_url != self.type_url:
      message = "Unpacking value of type '%s' to message of type '%s'."
      logging.warning(message, self.type_url, cls_type_url)

    return cls.FromSerializedBytes(self.value)


def TypeURL(cls: type[_V]) -> str:
  if cls.protobuf is None:
    raise ValueError("protobuf must be set on cls.")
  return f"type.googleapis.com/{cls.protobuf.DESCRIPTOR.full_name}"



#!/usr/bin/env python
"""A module with functions for working with GRR packages."""

import importlib
import inspect
import logging
import os
import sys

import pkg_resources


def _GetPkgResources(package_name, filepath):
  """A wrapper for the `pkg_resource.resource_filename` function."""
  requirement = pkg_resources.Requirement.parse(package_name)
  try:
    return pkg_resources.resource_filename(requirement, filepath)
  except pkg_resources.DistributionNotFound:
    # It may be that the working set is not in sync (e.g. if sys.path was
    # manipulated). Try to reload it just in case.
    pkg_resources.working_set = pkg_resources.WorkingSet()
    try:
      return pkg_resources.resource_filename(requirement, filepath)
    except pkg_resources.DistributionNotFound:
      logging.error("Distribution %s not found. Is it installed?", package_name)
      return None


def ResourcePath(package_name, filepath):
  """Computes a path to the specified package resource.

  Args:
    package_name: A name of the package where the resource is located.
    filepath: A path to the resource relative to the package location.

  Returns:
    A path to the resource or `None` if the resource cannot be found.
  """
  # If we are running a pyinstaller-built binary we rely on the sys.prefix
  # code below and avoid running this which will generate confusing error
  # messages.
  if not getattr(sys, "frozen", None):
    target = _GetPkgResources(package_name, filepath)
    if target and os.access(target, os.R_OK):
      return target

  # Installing from wheel places data_files relative to sys.prefix and not
  # site-packages. If we can not find in site-packages, check sys.prefix
  # instead.
  # https://python-packaging-user-guide.readthedocs.io/en/latest/distributing/#data-files
  target = os.path.join(sys.prefix, filepath)
  if target and os.access(target, os.R_OK):
    return target

  return None


def ModulePath(module_name) -> str:
  """Computes a path to the specified module.

  Args:
    module_name: A name of the module to get the path for.

  Returns:
    A path to the specified module.

  Raises:
    ImportError: If specified module cannot be imported.
  """
  module = importlib.import_module(module_name)
  path = inspect.getfile(module)

  # In case of modules with want a path to the directory rather than to the
  # `__init__.py` file itself.
  if os.path.basename(path).startswith("__init__."):
    path = os.path.dirname(path)

  # Sometimes __file__ points at a .pyc file, when we really mean the .py.
  if path.endswith(".pyc"):
    path = path[:-4] + ".py"

  return path



#!/usr/bin/env python
"""Network-related client rdfvalues."""

import binascii
import ipaddress
import logging
from typing import Optional, Union

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.util import precondition
from grr_response_core.lib.util import text
from grr_response_proto import jobs_pb2
from grr_response_proto import sysinfo_pb2


class NetworkEndpoint(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.NetworkEndpoint


class NetworkConnection(rdf_structs.RDFProtoStruct):
  """Information about a single network connection."""

  protobuf = sysinfo_pb2.NetworkConnection
  rdf_deps = [
      NetworkEndpoint,
  ]


IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]


class NetworkAddress(rdf_structs.RDFProtoStruct):
  """A network address.

  We'd prefer to use socket.inet_pton and  inet_ntop here, but they aren't
  available on windows before python 3.4. So we use the older IPv4 functions for
  v4 addresses and our own pure python implementations for IPv6.
  """

  protobuf = jobs_pb2.NetworkAddress

  @classmethod
  def FromPackedBytes(cls, ip: bytes) -> "NetworkAddress":
    result = cls()
    result.packed_bytes = ip

    if len(ip) * 8 == 32:
      result.address_type = jobs_pb2.NetworkAddress.INET
    elif len(ip) * 8 == 128:
      result.address_type = jobs_pb2.NetworkAddress.INET6
    else:
      message = f"Unexpected IP address length: {len(ip)}"
      raise ValueError(message)

    return result

  @property
  def human_readable_address(self) -> str:
    addr = self.AsIPAddr()
    if addr is not None:
      return str(addr)
    else:
      return ""

  @human_readable_address.setter
  def human_readable_address(self, value: str) -> None:
    precondition.AssertType(value, str)
    addr = ipaddress.ip_address(value)

    if isinstance(addr, ipaddress.IPv6Address):
      self.address_type = NetworkAddress.Family.INET6
    elif isinstance(addr, ipaddress.IPv4Address):
      self.address_type = NetworkAddress.Family.INET
    else:
      message = "IP address parsed to an unexpected value: {}".format(addr)
      raise AssertionError(message)

    self.packed_bytes = addr.packed

  def AsIPAddr(self) -> Optional[IPAddress]:
    """Returns the IP as an `IPAddress` object (if packed bytes are defined)."""
    precondition.AssertOptionalType(self.packed_bytes, bytes)

    if self.packed_bytes is None:
      return None

    try:
      if self.address_type == NetworkAddress.Family.INET:
        return ipaddress.IPv4Address(self.packed_bytes)
      if self.address_type == NetworkAddress.Family.INET6:
        return ipaddress.IPv6Address(self.packed_bytes)
    except ipaddress.AddressValueError:
      hex_packed_bytes = text.Hexify(self.packed_bytes)
      logging.error(
          "AddressValueError for %s (%s)", hex_packed_bytes, self.address_type
      )
      raise

    message = "IP address has invalid type: {}".format(self.address_type)
    raise ValueError(message)


class DNSClientConfiguration(rdf_structs.RDFProtoStruct):
  """DNS client config."""

  protobuf = sysinfo_pb2.DNSClientConfiguration


class MacAddress(rdfvalue.RDFBytes):
  """A MAC address."""

  @property
  def human_readable_address(self) -> str:
    return text.Hexify(self._value)

  @classmethod
  def FromHumanReadableAddress(cls, string: str):
    precondition.AssertType(string, str)
    return cls(binascii.unhexlify(string.encode("ascii")))


class Interface(rdf_structs.RDFProtoStruct):
  """A network interface on the client system."""

  protobuf = jobs_pb2.Interface
  rdf_deps = [
      MacAddress,
      NetworkAddress,
  ]

  def GetIPAddresses(self):
    """Return a list of IP addresses."""
    results = []
    for address in self.addresses:
      human_readable_address = address.human_readable_address
      if human_readable_address is not None:
        results.append(human_readable_address)

    return results



#!/usr/bin/env python
"""AFF4 RDFValue implementations for client information.

This module contains the RDFValue implementations used to communicate with the
client.
"""

from collections.abc import Mapping
import logging
import platform
import re
import socket
import sys

import distro
import psutil

from grr_response_core.lib import rdfvalue
from grr_response_core.lib import type_info
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import client_network as rdf_client_network
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import jobs_pb2
from grr_response_proto import knowledge_base_pb2
from grr_response_proto import sysinfo_pb2

# ntop does not exist on Windows.
# pylint: disable=g-socket-inet-aton,g-socket-inet-ntoa

# We try to support PEP 425 style component names if possible. This makes it
# possible to have wheel as an optional dependency.
try:
  # pytype: disable=import-error
  from wheel import pep425tags  # pylint: disable=g-import-not-at-top
  # pytype: enable=import-error
except ImportError:
  pep425tags = None

FS_ENCODING = sys.getfilesystemencoding() or sys.getdefaultencoding()

_LOCALHOST = "localhost"
_LOCALHOST_LOCALDOMAIN = "localhost.localdomain"
_ARPA_SUFFIXES = ("ip6.arpa", "in-addr.arpa")


class ClientURN(rdfvalue.RDFURN):
  """A client urn has to have a specific form."""

  # Valid client urns must match this expression.
  CLIENT_ID_RE = re.compile(r"^(aff4:)?/?(?P<clientid>(c|C)\.[0-9a-fA-F]{16})$")

  def __init__(self, initializer=None):
    super().__init__(initializer)

    if self._value and not self.Validate(self._value):
      raise type_info.TypeValueError("Client urn malformed: %s" % initializer)

  @classmethod
  def _Normalize(cls, string):
    normalized = super()._Normalize(string.strip())

    if normalized:
      match = cls.CLIENT_ID_RE.match(normalized)
      if not match:
        raise type_info.TypeValueError(
            "Client URN '{!r} from initializer {!r} malformed".format(
                normalized, string
            )
        )

      clientid = match.group("clientid")
      clientid_correctcase = "".join(
          (clientid[0].upper(), clientid[1:].lower())
      )

      normalized = normalized.replace(clientid, clientid_correctcase, 1)
    return normalized

  @classmethod
  def Validate(cls, value):
    if value:
      return bool(cls.CLIENT_ID_RE.match(str(value)))

    return False

  def Add(self, path):
    """Add a relative stem to the current value and return a new RDFURN.

    Note that this returns an RDFURN, not a ClientURN since the resulting object
    would not pass validation.

    Args:
      path: A string containing a relative path.

    Returns:
       A new RDFURN that can be chained.

    Raises:
       ValueError: if the path component is not a string.
    """
    if not isinstance(path, str):
      raise ValueError("Only strings should be added to a URN.")

    return rdfvalue.RDFURN(utils.JoinPath(self._value, path))


class PackageRepository(rdf_structs.RDFProtoStruct):
  """Description of the configured repositories (Yum etc).

  Describes the configured software package repositories.
  """

  protobuf = sysinfo_pb2.PackageRepository


class ManagementAgent(rdf_structs.RDFProtoStruct):
  """Description of the running management agent (puppet etc).

  Describes the state, last run timestamp, and name of the management agent
  installed on the system.
  """

  protobuf = sysinfo_pb2.ManagementAgent
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class PwEntry(rdf_structs.RDFProtoStruct):
  """Information about password structures."""

  protobuf = knowledge_base_pb2.PwEntry


class Group(rdf_structs.RDFProtoStruct):
  """Information about system posix groups."""

  protobuf = knowledge_base_pb2.Group
  rdf_deps = [
      PwEntry,
  ]


class User(rdf_structs.RDFProtoStruct):
  """Information about the users."""

  protobuf = knowledge_base_pb2.User
  rdf_deps = [
      PwEntry,
      rdfvalue.RDFDatetime,
  ]

  def __init__(self, initializer=None, **kwargs):
    if isinstance(initializer, KnowledgeBaseUser):
      # KnowledgeBaseUser was renamed to User, the protos are identical. This
      # allows for backwards compatibility with clients returning KBUser
      # objects.
      # TODO(user): remove once all clients are newer than 3.0.7.1.
      initializer = User.FromSerializedBytes(initializer.SerializeToBytes())
    super().__init__(initializer=initializer, **kwargs)


class KnowledgeBaseUser(User):
  """Backwards compatibility for old clients.

  Linux client action EnumerateUsers previously returned KnowledgeBaseUser
  objects.
  """


class KnowledgeBase(rdf_structs.RDFProtoStruct):
  """Information about the system and users."""

  protobuf = knowledge_base_pb2.KnowledgeBase
  rdf_deps = [
      User,
  ]

  def MergeOrAddUser(self, kb_user: User) -> None:
    """Merge a user into existing users or add new if it doesn't exist.

    Args:
      kb_user: A User rdfvalue.
    """

    user = self.GetUser(
        sid=kb_user.sid, uid=kb_user.uid, username=kb_user.username
    )
    if not user:
      self.users.Append(kb_user)
    else:
      for key, val in kb_user.AsDict().items():
        user.Set(key, val)

  def GetUser(self, sid=None, uid=None, username=None):
    """Retrieve a User based on sid, uid or username.

    On windows we first get a SID and use it to find the username.  We want to
    avoid combining users with name collisions, which occur when local users
    have the same username as domain users (something like Admin is particularly
    common).  So if a SID is provided, don't also try to match by username.

    On linux we first get a username, then use this to find the UID, so we want
    to combine these records or we end up with multiple partially-filled user
    records.

    TODO(user): this won't work at all well with a query for uid=0 because
    that is also the default for User objects that don't have uid
    set.

    Args:
      sid: Windows user sid
      uid: Linux/Darwin user id
      username: string

    Returns:
      rdf_client.User or None
    """
    if sid:
      for user in self.users:
        if user.sid == sid:
          return user
      return None
    if uid:
      for user in self.users:
        if user.uid == uid:
          return user
    if username:
      for user in self.users:
        if user.username == username:
          # Make sure we aren't combining different uids if we know them
          # user.uid = 0 is the default, which makes this more complicated.
          if uid and user.uid and user.uid != uid:
            return None
          else:
            return user

  def GetKbFieldNames(self):
    fields = set(self.type_infos.descriptor_names)

    # DEPRECATED_users field is maintained for backwards compatibility reasons
    # only, as type of the "users" field has changed from jobs.User to
    # knowledge_base.User.
    fields.remove("DEPRECATED_users")

    # We used to have a "hostname" field that contained an fqdn and is
    # therefore now renamed to "fqdn".
    # TODO(amoser): Remove once the artifact has removed the provides
    # section upstream.
    fields.add("hostname")

    fields.remove("users")
    for field in self.users.type_descriptor.type.type_infos.descriptor_names:
      fields.add("users.%s" % field)
    return sorted(fields)


class HardwareInfo(rdf_structs.RDFProtoStruct):
  """Various hardware information."""

  protobuf = sysinfo_pb2.HardwareInfo


class ClientInformation(rdf_structs.RDFProtoStruct):
  """The GRR client information."""

  protobuf = jobs_pb2.ClientInformation


class BufferReference(rdf_structs.RDFProtoStruct):
  """Stores information about a buffer in a file on the client."""

  protobuf = jobs_pb2.BufferReference
  rdf_deps = [
      rdf_paths.PathSpec,
  ]

  def __eq__(self, other):
    return self.data == other


class Process(rdf_structs.RDFProtoStruct):
  """Represent a process on the client."""

  protobuf = sysinfo_pb2.Process
  rdf_deps = [
      rdf_client_network.NetworkConnection,
  ]

  @classmethod
  def FromPsutilProcess(cls, psutil_process):
    response = cls()
    process_fields = ["pid", "ppid", "name", "exe", "username", "terminal"]

    with psutil_process.oneshot():
      for field in process_fields:
        try:
          value = getattr(psutil_process, field)
          if value is None:
            continue

          if callable(value):
            value = value()

          if not isinstance(value, int):
            value = utils.SmartUnicode(value)

          setattr(response, field, value)
        except (psutil.NoSuchProcess, psutil.AccessDenied, AttributeError):
          pass

      try:
        response.cmdline = list(psutil_process.cmdline())
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass
      except Exception as e:  # pylint: disable=broad-except
        # Windows now has Virtual Secure Mode (VSM) processes that have
        # additional memory protection. For those, cmdline() and cwd() will
        # raise a Windows Error 998 (ERROR_NOACCESS, Invalid access to memory
        # location).
        if not hasattr(e, "winerror") or e.winerror != 998:  # pytype: disable=attribute-error
          raise

      try:
        response.nice = psutil_process.nice()
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass

      try:
        # Not available on Windows.
        if hasattr(psutil_process, "uids"):
          (response.real_uid, response.effective_uid, response.saved_uid) = (
              psutil_process.uids()
          )
          (response.real_gid, response.effective_gid, response.saved_gid) = (
              psutil_process.gids()
          )
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass

      try:
        response.ctime = int(psutil_process.create_time() * 1e6)
        response.status = str(psutil_process.status())
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass

      try:
        # Not available on OSX.
        if hasattr(psutil_process, "cwd"):
          response.cwd = utils.SmartUnicode(psutil_process.cwd())
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass
      # We have seen psutil on macos raise OSError here with errno 2 - ENOENT -
      # but we haven't managed to reproduce this behavior. This might be fixed
      # in more recent versions of psutil but we play it safe and just catch the
      # error.
      except OSError:
        pass
      except Exception as e:  # pylint: disable=broad-except
        # Windows now has Virtual Secure Mode (VSM) processes that have
        # additional memory protection. For those, cmdline() and cwd() will
        # raise a Windows Error 998 (ERROR_NOACCESS, Invalid access to memory
        # location).
        if not hasattr(e, "winerror") or e.winerror != 998:  # pytype: disable=attribute-error
          raise

      try:
        response.num_threads = psutil_process.num_threads()
      except (psutil.NoSuchProcess, psutil.AccessDenied, RuntimeError):
        pass

      try:
        cpu_times = psutil_process.cpu_times()
        response.user_cpu_time = cpu_times.user
        response.system_cpu_time = cpu_times.system
        # psutil_process.get_cpu_percent() is very time consuming so we do not
        # collect it.
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass

      try:
        pmem = psutil_process.memory_info()
        response.RSS_size = pmem.rss  # pylint: disable=invalid-name
        response.VMS_size = pmem.vms  # pylint: disable=invalid-name
        response.memory_percent = psutil_process.memory_percent()
      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass

      # Due to a bug in psutil, this function is disabled for now
      # (https://github.com/giampaolo/psutil/issues/340)
      # try:
      #  for f in psutil_process.open_files():
      #    response.open_files.append(utils.SmartUnicode(f.path))
      # except (psutil.NoSuchProcess, psutil.AccessDenied):
      #  pass

      try:
        for c in psutil_process.connections():
          conn = response.connections.Append(
              family=c.family, type=c.type, pid=psutil_process.pid
          )

          try:
            conn.state = c.status
          except ValueError:
            logging.info(
                "Encountered unknown connection status (%s).", c.status
            )

          try:
            conn.local_address.ip, conn.local_address.port = c.laddr

            # Could be in state LISTEN.
            if c.raddr:
              conn.remote_address.ip, conn.remote_address.port = c.raddr
          except AttributeError:
            conn.local_address.ip, conn.local_address.port = c.local_address

            # Could be in state LISTEN.
            if c.remote_address:
              (conn.remote_address.ip, conn.remote_address.port) = (
                  c.remote_address
              )

      except (psutil.NoSuchProcess, psutil.AccessDenied):
        pass

      return response


class NamedPipe(rdf_structs.RDFProtoStruct):
  """An RDF wrapper for the named pipe message."""

  protobuf = sysinfo_pb2.NamedPipe
  rdf_deps = []


class SoftwarePackage(rdf_structs.RDFProtoStruct):
  """Represent an installed package on the client."""

  protobuf = sysinfo_pb2.SoftwarePackage

  @classmethod
  def Installed(cls, **kwargs):
    return SoftwarePackage(
        install_state=SoftwarePackage.InstallState.INSTALLED, **kwargs
    )

  @classmethod
  def Pending(cls, **kwargs):
    return SoftwarePackage(
        install_state=SoftwarePackage.InstallState.PENDING, **kwargs
    )

  @classmethod
  def Uninstalled(cls, **kwargs):
    return SoftwarePackage(
        install_state=SoftwarePackage.InstallState.UNINSTALLED, **kwargs
    )


class SoftwarePackages(rdf_structs.RDFProtoStruct):
  """A list of installed packages on the system."""

  protobuf = sysinfo_pb2.SoftwarePackages

  rdf_deps = [
      SoftwarePackage,
  ]


class LogMessage(rdf_structs.RDFProtoStruct):
  """A log message sent from the client to the server."""

  protobuf = jobs_pb2.LogMessage


class Uname(rdf_structs.RDFProtoStruct):
  """A protobuf to represent the current system."""

  protobuf = jobs_pb2.Uname
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]

  @property
  def arch(self):
    """Return a more standard representation of the architecture."""
    if self.machine in ["x86_64", "AMD64", "i686"]:
      # 32 bit binaries running on AMD64 will still have a i386 arch.
      if self.architecture == "32bit":
        return "i386"

      return "amd64"
    elif self.machine == "x86":
      return "i386"

    return self.machine

  def signature(self):
    """Returns a unique string that encapsulates the architecture."""
    # If the protobuf contains a proper pep425 tag return that.
    result = self.pep425tag
    if result:
      return result

    raise ValueError(
        "PEP 425 Signature not set - this is likely an old "
        "component file, please back it up and remove it."
    )

  @classmethod
  def FromCurrentSystem(cls):
    """Fill a Uname from the currently running platform."""
    uname = platform.uname()
    fqdn = socket.getfqdn()
    if (
        fqdn == _LOCALHOST
        or fqdn == _LOCALHOST_LOCALDOMAIN
        or any(fqdn.endswith(suffix) for suffix in _ARPA_SUFFIXES)
    ):
      # Avoid returning 'localhost' when there is a better value to use.
      fqdn = socket.gethostname()
    system = uname[0]
    architecture, _ = platform.architecture()
    if system == "Windows":
      service_pack = platform.win32_ver()[2]
      kernel = uname[3]  # 5.1.2600
      release = uname[2]  # XP, 2000, 7
      version = uname[3] + service_pack  # 5.1.2600 SP3, 6.1.7601 SP1
    elif system == "Darwin":
      kernel = uname[2]  # 12.2.0
      release = "OSX"  # OSX
      version = platform.mac_ver()[0]  # 10.8.2
    elif system == "Linux":
      kernel = uname[2]  # 3.2.5
      release = distro.name()
      version = distro.version()

    # Emulate PEP 425 naming conventions - e.g. cp27-cp27mu-linux_x86_64.
    if pep425tags:
      try:
        # 0.33.6
        pep_platform = pep425tags.get_platform()
      except TypeError:
        # 0.34.2
        pep_platform = pep425tags.get_platform(None)
      pep425tag = "%s%s-%s-%s" % (
          pep425tags.get_abbr_impl(),
          pep425tags.get_impl_ver(),
          str(pep425tags.get_abi_tag()).lower(),
          pep_platform,
      )
    else:
      # For example: windows_7_amd64
      pep425tag = "%s_%s_%s" % (system, release, architecture)

    return cls(
        system=system,
        architecture=architecture,
        release=release,
        version=version,
        machine=uname[4],  # x86, x86_64
        kernel=kernel,
        fqdn=fqdn,
        pep425tag=pep425tag,
    )


class StartupInfo(rdf_structs.RDFProtoStruct):
  """Information about the startup of a GRR agent."""

  protobuf = jobs_pb2.StartupInfo
  rdf_deps = [
      ClientInformation,
      rdfvalue.RDFDatetime,
  ]


class OSXServiceInformation(rdf_structs.RDFProtoStruct):
  """OSX Service (launchagent/daemon)."""

  protobuf = sysinfo_pb2.OSXServiceInformation
  rdf_deps = [
      rdfvalue.RDFURN,
  ]


# Start of the Registry Specific Data types
class RunKey(rdf_structs.RDFProtoStruct):
  protobuf = sysinfo_pb2.RunKey


class ClientCrash(rdf_structs.RDFProtoStruct):
  """Details of a client crash."""

  protobuf = jobs_pb2.ClientCrash
  rdf_deps = [
      ClientInformation,
      ClientURN,
      rdfvalue.RDFDatetime,
      rdfvalue.SessionID,
  ]


class EdrAgent(rdf_structs.RDFProtoStruct):
  """An RDF wrapper for protobuf message containing EDR agent metadata."""

  protobuf = jobs_pb2.EdrAgent
  rdf_deps = []


class FleetspeakValidationInfoTag(rdf_structs.RDFProtoStruct):
  """Dictionary entry in FleetspeakValidationInfo."""

  protobuf = jobs_pb2.FleetspeakValidationInfoTag


class FleetspeakValidationInfo(rdf_structs.RDFProtoStruct):
  """Dictionary-like struct containing Fleetspeak ValidationInfo."""

  protobuf = jobs_pb2.FleetspeakValidationInfo
  rdf_deps = [FleetspeakValidationInfoTag]

  @classmethod
  def FromStringDict(cls, dct: Mapping[str, str]) -> "FleetspeakValidationInfo":
    instance = cls()
    for key, value in dct.items():
      instance.tags.Append(key=key, value=value)
    return instance

  def ToStringDict(self) -> Mapping[str, str]:
    return {tag.key: tag.value for tag in self.tags}


class ClientSummary(rdf_structs.RDFProtoStruct):
  """Object containing client's summary data."""

  protobuf = jobs_pb2.ClientSummary
  rdf_deps = [
      ClientInformation,
      ClientURN,
      EdrAgent,
      rdf_client_network.Interface,
      rdfvalue.RDFDatetime,
      Uname,
      User,
      FleetspeakValidationInfo,
  ]


class VersionString(rdfvalue.RDFString):

  @property
  def versions(self):
    result = []
    for x in str(self).split("."):
      try:
        result.append(int(x))
      except ValueError:
        break

    return result



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import flows as rdf_flows
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2


def ToProtoGrrMessage(rdf: rdf_flows.GrrMessage) -> jobs_pb2.GrrMessage:
  return rdf.AsPrimitiveProto()


def ToRDFGrrMessage(proto: jobs_pb2.GrrMessage) -> rdf_flows.GrrMessage:
  return rdf_flows.GrrMessage.FromSerializedBytes(proto.SerializeToString())


def ToProtoGrrStatus(rdf: rdf_flows.GrrStatus) -> jobs_pb2.GrrStatus:
  return rdf.AsPrimitiveProto()


def ToRDFGrrStatus(proto: jobs_pb2.GrrStatus) -> rdf_flows.GrrStatus:
  return rdf_flows.GrrStatus.FromSerializedBytes(proto.SerializeToString())


def ToProtoFlowProcessingRequest(
    rdf: rdf_flows.FlowProcessingRequest,
) -> flows_pb2.FlowProcessingRequest:
  return rdf.AsPrimitiveProto()


def ToRDFFlowProcessingRequest(
    proto: flows_pb2.FlowProcessingRequest,
) -> rdf_flows.FlowProcessingRequest:
  return rdf_flows.FlowProcessingRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoNotification(rdf: rdf_flows.Notification) -> jobs_pb2.Notification:
  return rdf.AsPrimitiveProto()


def ToRDFNotification(proto: jobs_pb2.Notification) -> rdf_flows.Notification:
  return rdf_flows.Notification.FromSerializedBytes(proto.SerializeToString())


def ToProtoFlowNotification(
    rdf: rdf_flows.FlowNotification,
) -> jobs_pb2.FlowNotification:
  return rdf.AsPrimitiveProto()


def ToRDFFlowNotification(
    proto: jobs_pb2.FlowNotification,
) -> rdf_flows.FlowNotification:
  return rdf_flows.FlowNotification.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPackedMessageList(
    rdf: rdf_flows.PackedMessageList,
) -> jobs_pb2.PackedMessageList:
  return rdf.AsPrimitiveProto()


def ToRDFPackedMessageList(
    proto: jobs_pb2.PackedMessageList,
) -> rdf_flows.PackedMessageList:
  return rdf_flows.PackedMessageList.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoMessageList(rdf: rdf_flows.MessageList) -> jobs_pb2.MessageList:
  return rdf.AsPrimitiveProto()


def ToRDFMessageList(proto: jobs_pb2.MessageList) -> rdf_flows.MessageList:
  return rdf_flows.MessageList.FromSerializedBytes(proto.SerializeToString())


def ToProtoCipherProperties(
    rdf: rdf_flows.CipherProperties,
) -> jobs_pb2.CipherProperties:
  return rdf.AsPrimitiveProto()


def ToRDFCipherProperties(
    proto: jobs_pb2.CipherProperties,
) -> rdf_flows.CipherProperties:
  return rdf_flows.CipherProperties.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCipherMetadata(
    rdf: rdf_flows.CipherMetadata,
) -> jobs_pb2.CipherMetadata:
  return rdf.AsPrimitiveProto()


def ToRDFCipherMetadata(
    proto: jobs_pb2.CipherMetadata,
) -> rdf_flows.CipherMetadata:
  return rdf_flows.CipherMetadata.FromSerializedBytes(proto.SerializeToString())


def ToProtoFlowLog(rdf: rdf_flows.FlowLog) -> jobs_pb2.FlowLog:
  return rdf.AsPrimitiveProto()


def ToRDFFlowLog(proto: jobs_pb2.FlowLog) -> rdf_flows.FlowLog:
  return rdf_flows.FlowLog.FromSerializedBytes(proto.SerializeToString())


def ToProtoHttpRequest(rdf: rdf_flows.HttpRequest) -> jobs_pb2.HttpRequest:
  return rdf.AsPrimitiveProto()


def ToRDFHttpRequest(proto: jobs_pb2.HttpRequest) -> rdf_flows.HttpRequest:
  return rdf_flows.HttpRequest.FromSerializedBytes(proto.SerializeToString())


def ToProtoClientCommunication(
    rdf: rdf_flows.ClientCommunication,
) -> jobs_pb2.ClientCommunication:
  return rdf.AsPrimitiveProto()


def ToRDFClientCommunication(
    proto: jobs_pb2.ClientCommunication,
) -> rdf_flows.ClientCommunication:
  return rdf_flows.ClientCommunication.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoEmptyFlowArgs(
    rdf: rdf_flows.EmptyFlowArgs,
) -> flows_pb2.EmptyFlowArgs:
  return rdf.AsPrimitiveProto()


def ToRDFEmptyFlowArgs(
    proto: flows_pb2.EmptyFlowArgs,
) -> rdf_flows.EmptyFlowArgs:
  return rdf_flows.EmptyFlowArgs.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""Configuration parameters for the test subsystem."""

from grr_response_core.lib import config_lib

config_lib.DEFINE_constant_string(
    "Test.data_dir",
    default="%(grr_response_test/test_data@grr-response-test|resource)",
    help="The directory where test data exist.")

config_lib.DEFINE_constant_string(
    "Test.additional_test_config",
    default="%(Test.data_dir)/localtest.yaml",
    help="The path to a test config with local customizations.")

config_lib.DEFINE_integer("SharedMemoryDB.port", 0,
                          "Port used to connect to SharedMemoryDB server.")

config_lib.DEFINE_string(
    "Mysql.schema_dump_path", "%(grr_response_server/databases/mysql.ddl@"
    "grr-response-server|resource)",
    "Location of the dumped MySQL schema path.")



#!/usr/bin/env python
"""Contains a stats-collector singleton shared across a GRR process."""

import logging
import threading

from grr_response_core.stats import stats_collector

_stats_singleton = None
_init_lock = threading.Lock()
_metadatas = []


class StatsNotInitializedError(Exception):

  def __init__(self):
    super().__init__("No StatsCollector has been initialized yet.")


def Set(collector: stats_collector.StatsCollector):
  """Initializes the StatsCollector singleton and registers metrics with it."""
  global _stats_singleton

  with _init_lock:
    if _stats_singleton is None:
      _stats_singleton = collector
      for metadata in _metadatas:
        _stats_singleton.RegisterMetric(metadata)
    else:
      # TODO(user): Throw an exception instead, once it is confirmed that it
      # is ok to do so.
      logging.warning("Tried to re-initialize global stats collector.")


def Get() -> stats_collector.StatsCollector:
  """Returns an initialized stats-collector.

  Raises:
    StatsNotInitializedError: If no stats-collector has been initialized yet.
  """
  if _stats_singleton is None:
    raise StatsNotInitializedError()
  return _stats_singleton


def RegisterMetric(metadata):
  """Registers a Metric with the StatsCollector."""
  with _init_lock:
    _metadatas.append(metadata)
    if _stats_singleton is not None:
      _stats_singleton.RegisterMetric(metadata)



#!/usr/bin/env python
"""Windows paths detection classes."""

import re

from grr_response_core.path_detection import core


class RunDllExtractor(core.Extractor):
  """Extractor that extracts rundll paths."""

  def Extract(self, components):
    """Extracts interesting paths from a given path.

    Args:
      components: Source string represented as a list of components.

    Returns:
      A list of extracted paths (as strings).
    """

    rundll_index = -1
    for index, component in enumerate(components):
      if component.lower().endswith("rundll32.exe"):
        rundll_index = index

    if rundll_index == -1:
      return []

    components = components[(rundll_index + 1) :]

    # We expect components after "rundll32.exe" to point at a DLL and a
    # function. For example:
    # rundll32.exe "C:\Windows\system32\advpack.dll",DelNodeRunDLL32
    last_component = components[-1].rsplit(",", 1)[0]

    extracted_path = " ".join(components[0:-1] + [last_component])
    return [extracted_path]


class ExecutableExtractor(core.Extractor):
  """Extractor for ordinary paths."""

  EXECUTABLE_EXTENSIONS = (
      "exe",
      "com",
      "bat",
      "dll",
      "msi",
      "sys",
      "scr",
      "pif",
  )

  def Extract(self, components):
    """Extracts interesting paths from a given path.

    Args:
      components: Source string represented as a list of components.

    Returns:
      A list of extracted paths (as strings).
    """

    for index, component in enumerate(components):
      if component.lower().endswith(self.EXECUTABLE_EXTENSIONS):
        extracted_path = " ".join(components[0 : index + 1])
        return [extracted_path]

    return []


class EnvVarsPostProcessor(core.PostProcessor):
  """PostProcessor that replaces env variables with predefined values."""

  # Service keys have peculiar ways of specifying systemroot, these regexes
  # will be used to convert them to standard expansions.
  SYSTEMROOT_RE = re.compile(r"^\\SystemRoot", flags=re.IGNORECASE)
  SYSTEM32_RE = re.compile(r"^system32", flags=re.IGNORECASE)

  # Regex that matches Windows Registry environment variables.
  WIN_ENVIRON_REGEX = re.compile(r"%([^%]+?)%")

  def __init__(self, vars_map):
    """EnvVarsPostProcessor constructor.

    Args:
      vars_map: Dictionary of "string" -> "string|list", i.e. a mapping of
        environment variables names to their suggested values or to lists of
        their suggested values.
    """
    super(core.PostProcessor, self).__init__()

    self.vars_map = {}
    for var_name, value in vars_map.items():
      var_regex = re.compile(
          re.escape("%" + var_name + "%"), flags=re.IGNORECASE
      )
      self.vars_map[var_name.lower()] = (var_regex, value)

  def Process(self, path):
    """Processes a given path.

    Args:
      path: Path (as a string) to post-process.

    Returns:
      A list of paths with environment variables replaced with their
      values. If the mapping had a list of values for a particular variable,
      instead of just one value, then all possible replacements will be
      returned.
    """
    path = re.sub(self.SYSTEMROOT_RE, r"%systemroot%", path, count=1)
    path = re.sub(self.SYSTEM32_RE, r"%systemroot%\\system32", path, count=1)

    matches_iter = self.WIN_ENVIRON_REGEX.finditer(path)
    var_names = set(m.group(1).lower() for m in matches_iter)

    results = [path]
    for var_name in var_names:
      try:
        var_regex, var_value = self.vars_map[var_name]
      except KeyError:
        continue

      if isinstance(var_value, str):
        replacements = [var_value]
      else:
        replacements = var_value

      processed_results = []
      for result in results:
        for repl in replacements:
          # Using lambda here, as otherwise Python interprets \\f as a
          # backreference (same applies to \\0 and \\1). When using a
          # function as a replacement argument, backreferences are ignored.
          # pylint: disable=cell-var-from-loop
          processed_results.append(var_regex.sub(lambda _: repl, result))
      results = processed_results

    return results


def CreateWindowsRegistryExecutablePathsDetector(vars_map=None):
  """Creates Windows paths detector.

  Commandline strings can be space separated and contain options.
  e.g. C:\\Program Files\\ACME Corporation\\wiz.exe /quiet /blah

  See here for microsoft doco on commandline parsing:
  http://msdn.microsoft.com/en-us/library/windows/desktop/ms682425(v=vs.85).aspx

  Args:
    vars_map: Environment variables mapping. Default is None.

  Returns:
    A detector (core.Detector instance).
  """
  return core.Detector(
      extractors=[RunDllExtractor(), ExecutableExtractor()],
      post_processors=[EnvVarsPostProcessor(vars_map or {})],
  )


def DetectExecutablePaths(source_values, vars_map=None):
  """Detects paths in a list of Windows Registry strings.

  Args:
    source_values: A list of strings to detect paths in.
    vars_map: Dictionary of "string" -> "string|list", i.e. a mapping of
      environment variables names to their suggested values or to lists of their
      suggested values.

  Yields:
    A list of detected paths (as strings).
  """
  detector = CreateWindowsRegistryExecutablePathsDetector(vars_map=vars_map)

  for source_value in source_values:
    for result in detector.Detect(source_value):
      yield result



#!/usr/bin/env python
"""GRR config parsing code."""

import abc
import configparser
import errno
import io
import logging
import os
from typing import Any, BinaryIO

import yaml


class Error(Exception):
  pass


class SaveDataError(Error):
  """Raised when config data can't be saved."""


class SaveDataPathNotSpecifiedError(SaveDataError):
  """Raised on save, if config path is not specified."""


class ReadDataError(Error):
  """Raised when config data can't be read."""


class ReadDataPathNotSpecifiedError(ReadDataError):
  """Raised on read, if config path is not specified."""


class ReadDataPermissionError(ReadDataError):
  """Raised when config is present but can't be accessed."""


class GRRConfigParser(metaclass=abc.ABCMeta):
  """The base class for all GRR configuration parsers."""

  def __init__(self, config_path: str) -> None:
    self._config_path = config_path

  def __str__(self) -> str:
    return '<%s config_path="%s">' % (
        self.__class__.__name__,
        self._config_path,
    )

  @property
  def config_path(self) -> str:
    return self._config_path

  @abc.abstractmethod
  def Copy(self) -> "GRRConfigParser":
    raise NotImplementedError()

  @abc.abstractmethod
  def SaveData(self, raw_data: dict[str, Any]) -> None:
    raise NotImplementedError()

  @abc.abstractmethod
  def ReadData(self) -> dict[str, Any]:
    """Convert the file to a more suitable data structure.

    Returns:
    The standard data format from this method is for example:

    {
     name: default_value;
     name2: default_value2;

     "Context1": {
         name: value,
         name2: value,

         "Nested Context" : {
           name: value;
         };
      },
     "Context2": {
         name: value,
      }
    }

    i.e. raw_data is an OrderedDict() with keys representing parameter names
    and values representing values. Contexts are represented by nested
    OrderedDict() structures with similar format.

    Note that support for contexts is optional and depends on the config file
    format. If contexts are not supported, a flat OrderedDict() is returned.
    """
    raise NotImplementedError()


class GRRConfigFileParser(GRRConfigParser):
  """Base class for file-based parsers."""

  @abc.abstractmethod
  def RawDataToBytes(self, raw_data: dict[str, Any]) -> bytes:
    raise NotImplementedError()

  @abc.abstractmethod
  def RawDataFromBytes(self, b: bytes) -> dict[str, Any]:
    raise NotImplementedError()

  def SaveDataToFD(
      self, raw_data: dict[str, Any], fd: io.BufferedWriter
  ) -> None:
    fd.write(self.RawDataToBytes(raw_data))

  def ReadDataFromFD(self, fd: BinaryIO) -> dict[str, Any]:
    return self.RawDataFromBytes(fd.read())

  def SaveData(self, raw_data: dict[str, Any]) -> None:
    """Store the raw data as our configuration."""
    if not self.config_path:
      raise SaveDataPathNotSpecifiedError("Parser's config_path is empty.")

    logging.info("Writing back configuration to file %s", self.config_path)
    # Ensure intermediate directories exist
    try:
      os.makedirs(os.path.dirname(self.config_path))
    except OSError:
      pass

    try:
      # We can not use the standard open() call because we need to
      # enforce restrictive file permissions on the created file.
      mode = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
      fd = os.open(self.config_path, mode, 0o600)
      with os.fdopen(fd, "wb") as config_file:
        config_file.write(self.RawDataToBytes(raw_data))

    except OSError as e:
      logging.exception(
          "Unable to write config file %s: %s.", self.config_path, e
      )
      raise SaveDataError(
          f"Unable to write config file {self.config_path}: {e}."
      ) from e

  def ReadData(self) -> dict[str, Any]:
    if not self.config_path:
      raise ReadDataPathNotSpecifiedError("Parser's config_path is empty.")

    try:
      # TODO(user): a normal "open" would do, but we have a test
      # in config_lib_test.py that relies on mocking io.open.
      with io.open(self.config_path, "rb") as fd:
        return self.ReadDataFromFD(fd)
    except OSError as e:
      if e.errno == errno.EACCES:
        # Specifically catch access denied errors, this usually indicates the
        # user wanted to read the file, and it existed, but they lacked the
        # permissions.
        raise ReadDataPermissionError(e) from e

      return dict()

  def Copy(self) -> "GRRConfigFileParser":
    return self.__class__(self._config_path)  # pytype: disable=not-instantiable


class IniConfigFileParser(GRRConfigFileParser):
  """A parser for ini style config files."""

  def RawDataToBytes(self, raw_data: dict[str, Any]) -> bytes:
    parser = self._Parser()

    for key, value in raw_data.items():
      parser.set("", key, value=value)

    sio = io.StringIO()
    parser.write(sio)
    return sio.getvalue().encode("utf-8")

  def RawDataFromBytes(self, b: bytes) -> dict[str, Any]:
    parser = self._Parser()
    parser.read_file(b.decode("utf-8").splitlines())

    raw_data = dict()
    for section in parser.sections():
      for key, value in parser.items(section):
        raw_data[".".join([section, key])] = value

    return raw_data

  def _Parser(self):
    parser = configparser.RawConfigParser()
    parser.optionxform = str
    return parser


class YamlConfigFileParser(GRRConfigFileParser):
  """A parser for yaml style config files."""

  def RawDataToBytes(self, raw_data: dict[str, Any]) -> bytes:
    return yaml.safe_dump(raw_data).encode("utf-8")

  def RawDataFromBytes(self, b: bytes) -> dict[str, Any]:
    return yaml.safe_load(b.decode("utf-8")) or dict()


class FileParserDataWrapper(GRRConfigParser):
  """Wrapper that makes GRRConfigFileParser read data from predefined bytes."""

  def __init__(self, data: bytes, parser: GRRConfigFileParser):
    # Use an empty config_path.
    super().__init__("")
    self._data = data
    self._parser = parser

  def SaveData(self, raw_data: dict[str, Any]) -> None:
    raise SaveDataError("File parser initialized from bytes can't save data.")

  def ReadData(self) -> dict[str, Any]:
    return self._parser.RawDataFromBytes(self._data)

  def Copy(self) -> "FileParserDataWrapper":
    return self.__class__(self._data, self._parser)


_ADDITIONAL_PARSERS: dict[str, type[GRRConfigParser]] = {}


def RegisterParserClass(scheme: str, parser_cls: type[GRRConfigParser]) -> None:
  _ADDITIONAL_PARSERS[scheme] = parser_cls


def _GetParserClassFromPath(path: str) -> type[GRRConfigParser]:
  """Returns the appropriate parser class from the path."""
  # Find the configuration parser.
  path_scheme = path.split("://")[0]
  for scheme, parser_cls in _ADDITIONAL_PARSERS.items():
    if scheme == path_scheme:
      return parser_cls

  # Handle the filename.
  extension = os.path.splitext(path)[1]
  if extension in [".yaml", ".yml"]:
    return YamlConfigFileParser

  return IniConfigFileParser


def GetParserFromPath(path: str) -> GRRConfigParser:
  """Returns the appropriate parser class from the path."""
  cls = _GetParserClassFromPath(path)
  return cls(path)  # pytype: disable=not-instantiable



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import read_low_level as rdf_read_low_level
from grr_response_proto import read_low_level_pb2


def ToProtoReadLowLevelArgs(
    rdf: rdf_read_low_level.ReadLowLevelArgs,
) -> read_low_level_pb2.ReadLowLevelArgs:
  return rdf.AsPrimitiveProto()


def ToRDFReadLowLevelArgs(
    proto: read_low_level_pb2.ReadLowLevelArgs,
) -> rdf_read_low_level.ReadLowLevelArgs:
  return rdf_read_low_level.ReadLowLevelArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoReadLowLevelFlowResult(
    rdf: rdf_read_low_level.ReadLowLevelFlowResult,
) -> read_low_level_pb2.ReadLowLevelFlowResult:
  return rdf.AsPrimitiveProto()


def ToRDFReadLowLevelFlowResult(
    proto: read_low_level_pb2.ReadLowLevelFlowResult,
) -> rdf_read_low_level.ReadLowLevelFlowResult:
  return rdf_read_low_level.ReadLowLevelFlowResult.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoReadLowLevelRequest(
    rdf: rdf_read_low_level.ReadLowLevelRequest,
) -> read_low_level_pb2.ReadLowLevelRequest:
  return rdf.AsPrimitiveProto()


def ToRDFReadLowLevelRequest(
    proto: read_low_level_pb2.ReadLowLevelRequest,
) -> rdf_read_low_level.ReadLowLevelRequest:
  return rdf_read_low_level.ReadLowLevelRequest.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoReadLowLevelResult(
    rdf: rdf_read_low_level.ReadLowLevelResult,
) -> read_low_level_pb2.ReadLowLevelResult:
  return rdf.AsPrimitiveProto()


def ToRDFReadLowLevelResult(
    proto: read_low_level_pb2.ReadLowLevelResult,
) -> rdf_read_low_level.ReadLowLevelResult:
  return rdf_read_low_level.ReadLowLevelResult.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""Default implementation for a stats-collector."""

import abc

from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import stats as rdf_stats
from grr_response_core.stats import stats_collector
from grr_response_core.stats import stats_utils


def _FieldsToKey(fields):
  """Converts a list of field values to a metric key."""
  return tuple(fields) if fields else ()


class _Metric(metaclass=abc.ABCMeta):
  """Base class for all the metric objects used by the DefaultStatsCollector.

  See stats_collector for more info.
  """

  def __init__(self, field_defs):
    """Initializes the metric.

    Args:
      field_defs: A list of (field-name, field-type) tuples describing the
        dimensions for the metric.
    """
    self._field_defs = field_defs
    self._metric_values = {}

  @abc.abstractmethod
  def _DefaultValue(self):
    """Returns the default value of a metric.

    For counters, the default value is 0, for event metrics, the default
    value is a distribution, and for gauges, the default value is 0, 0.0 or
    the empty string depending on the type of the gauge (int, float or str).
    """

  def Get(self, fields=None):
    """Gets the metric value corresponding to the given field values."""
    if not self._field_defs and fields:
      raise ValueError(
          "Metric was registered without fields, "
          "but following fields were provided: %s." % (fields,)
      )

    if self._field_defs and not fields:
      raise ValueError(
          "Metric was registered with fields (%s), but no fields were provided."
          % self._field_defs
      )

    if self._field_defs and fields and len(self._field_defs) != len(fields):
      raise ValueError(
          "Metric was registered with %d fields (%s), but "
          "%d fields were provided (%s)."
          % (len(self._field_defs), self._field_defs, len(fields), fields)
      )

    metric_value = self._metric_values.get(_FieldsToKey(fields))
    return self._DefaultValue() if metric_value is None else metric_value

  def ListFieldsValues(self):
    """Returns a list of tuples of all field values used with the metric."""
    return list(self._metric_values) if self._field_defs else []


class _CounterMetric(_Metric):
  """Simple counter metric (see stats_collector for more info)."""

  def _DefaultValue(self):
    return 0

  def Increment(self, delta, fields=None):
    """Increments counter value by a given delta."""
    if delta < 0:
      raise ValueError(
          "Counter increment should not be < 0 (received: %d)" % delta
      )

    self._metric_values[_FieldsToKey(fields)] = self.Get(fields=fields) + delta


class _EventMetric(_Metric):
  """_EventMetric provides detailed stats, like averages, distribution, etc.

  See stats_collector for more info.

  Args:
    bins: A list of numbers defining distribution buckets for the metric. If
      empty/None, a default list of buckets is used.
    fields: A list of (field-name, field-type) tuples describing the dimensions
      for the metric.
  """

  def __init__(self, bins, fields):
    super().__init__(fields)
    self._bins = bins or [
        0.0,
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.75,
        1,
        1.5,
        2,
        2.5,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        15,
        20,
        50,
        100,
    ]

  def _DefaultValue(self):
    return rdf_stats.Distribution(bins=self._bins)

  def Record(self, value, fields=None):
    """Records the given observation in a distribution."""
    key = _FieldsToKey(fields)
    metric_value = self._metric_values.get(key)
    if metric_value is None:
      metric_value = self._DefaultValue()
      self._metric_values[key] = metric_value
    metric_value.Record(value)


class _GaugeMetric(_Metric):
  """A metric whose value can increase or decrease.

  See stats_collector for more info.

  Args:
    value_type: Type of the gauge (one of int, str or float).
    fields: A list of (field-name, field-type) tuples describing the dimensions
      for the metric.
  """

  def __init__(self, value_type, fields):
    super().__init__(fields)
    self._value_type = value_type

  def _DefaultValue(self):
    return self._value_type()

  def Set(self, value, fields=None):
    """Sets the metric's current value."""
    self._metric_values[_FieldsToKey(fields)] = self._value_type(value)

  def SetCallback(self, callback, fields=None):
    """Attaches the given callback to the metric."""
    self._metric_values[_FieldsToKey(fields)] = callback

  def Get(self, fields=None):
    """Returns current metric's value (executing a callback if needed)."""
    result = super().Get(fields=fields)
    if callable(result):
      return result()
    else:
      return result


class DefaultStatsCollector(stats_collector.StatsCollector):
  """Default implementation for a stats-collector."""

  def __init__(self):
    self._counter_metrics = {}
    self._gauge_metrics = {}
    self._event_metrics = {}

    super().__init__()

  def _InitializeMetric(self, metadata):
    """See base class."""
    field_defs = stats_utils.FieldDefinitionTuplesFromProtos(
        metadata.fields_defs
    )
    if metadata.metric_type == rdf_stats.MetricMetadata.MetricType.COUNTER:
      self._counter_metrics[metadata.varname] = _CounterMetric(field_defs)
    elif metadata.metric_type == rdf_stats.MetricMetadata.MetricType.EVENT:
      self._event_metrics[metadata.varname] = _EventMetric(
          list(metadata.bins), field_defs
      )
    elif metadata.metric_type == rdf_stats.MetricMetadata.MetricType.GAUGE:
      value_type = stats_utils.PythonTypeFromMetricValueType(
          metadata.value_type
      )
      self._gauge_metrics[metadata.varname] = _GaugeMetric(
          value_type, field_defs
      )
    else:
      raise ValueError("Unknown metric type: %s." % metadata.metric_type)

  @utils.Synchronized
  def IncrementCounter(self, metric_name, delta=1, fields=None):
    """See base class."""
    if delta < 0:
      raise ValueError("Invalid increment for counter: %d." % delta)
    self._counter_metrics[metric_name].Increment(delta, fields)

  @utils.Synchronized
  def RecordEvent(self, metric_name, value, fields=None):
    """See base class."""
    self._event_metrics[metric_name].Record(value, fields)

  @utils.Synchronized
  def SetGaugeValue(self, metric_name, value, fields=None):
    """See base class."""
    self._gauge_metrics[metric_name].Set(value, fields)

  @utils.Synchronized
  def SetGaugeCallback(self, metric_name, callback, fields=None):
    """See base class."""
    self._gauge_metrics[metric_name].SetCallback(callback, fields)

  def GetMetricFields(self, metric_name):
    """See base class."""
    return self._GetMetric(metric_name).ListFieldsValues()

  def GetMetricValue(self, metric_name, fields=None):
    """See base class."""
    return self._GetMetric(metric_name).Get(fields)

  def _GetMetric(self, metric_name):
    """Fetches the metric object corresponding to the given name."""
    if metric_name in self._counter_metrics:
      return self._counter_metrics[metric_name]
    elif metric_name in self._event_metrics:
      return self._event_metrics[metric_name]
    elif metric_name in self._gauge_metrics:
      return self._gauge_metrics[metric_name]
    else:
      raise ValueError("Metric %s is not registered." % metric_name)



#!/usr/bin/env python
"""RDFValues used with Yara."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import flows_pb2


class YaraSignature(rdfvalue.RDFString):
  pass


class YaraSignatureShard(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.YaraSignatureShard
  rdf_deps = []


class YaraProcessScanRequest(rdf_structs.RDFProtoStruct):
  """Args for YaraProcessScan flow and client action."""

  protobuf = flows_pb2.YaraProcessScanRequest
  rdf_deps = [
      YaraSignature,
      YaraSignatureShard,
      rdfvalue.ByteSize,
      rdfvalue.Duration,
  ]

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    # These default values were migrated from the Protobuf definition.
    if not self.HasField("include_errors_in_results"):
      self.include_errors_in_results = (
          YaraProcessScanRequest.ErrorPolicy.NO_ERRORS
      )
    if not self.HasField("include_misses_in_results"):
      self.include_misses_in_results = False
    if not self.HasField("ignore_grr_process"):
      self.ignore_grr_process = True
    if not self.HasField("chunk_size"):
      self.chunk_size = 100 * 1024 * 1024  # 100 MiB
    if not self.HasField("overlap_size"):
      self.overlap_size = 10 * 1024 * 1024  # 10 MiB
    if not self.HasField("skip_special_regions"):
      self.skip_special_regions = False
    if not self.HasField("skip_mapped_files"):
      self.skip_mapped_files = True
    if not self.HasField("skip_shared_regions"):
      self.skip_shared_regions = False
    if not self.HasField("skip_executable_regions"):
      self.skip_executable_regions = False
    if not self.HasField("skip_readonly_regions"):
      self.skip_readonly_regions = False
    if not self.HasField("dump_process_on_match"):
      self.dump_process_on_match = False
    if not self.HasField("process_dump_size_limit"):
      self.process_dump_size_limit = 0
    if not self.HasField("context_window"):
      self.context_window = 50


class ProcessMemoryError(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.ProcessMemoryError
  rdf_deps = [rdf_client.Process]


class YaraStringMatch(rdf_structs.RDFProtoStruct):
  """A result of Yara string matching."""

  protobuf = flows_pb2.YaraStringMatch
  rdf_deps = []

  @classmethod
  def FromLibYaraStringMatch(
      cls, yara_string_match, yara_string_match_instance
  ):
    # Format is described in
    # https://yara.readthedocs.io/en/stable/yarapython.html#yara.StringMatch
    res = cls()
    res.offset = yara_string_match_instance.offset
    res.string_id = yara_string_match.identifier
    res.data = yara_string_match_instance.matched_data

    return res


class YaraMatch(rdf_structs.RDFProtoStruct):
  """A result of Yara matching."""

  protobuf = flows_pb2.YaraMatch
  rdf_deps = [YaraStringMatch]

  @classmethod
  def FromLibYaraMatch(cls, yara_match, data, context_window):
    string_matches = []
    for yara_string_match in yara_match.strings:
      for sm in yara_string_match.instances:
        result = YaraStringMatch.FromLibYaraStringMatch(yara_string_match, sm)
        if context_window > 0:
          context = data[
              sm.offset - context_window : sm.offset + context_window
          ]
          result.context = context
        string_matches.append(result)

    res = cls()
    res.rule_name = yara_match.rule
    res.string_matches = string_matches
    return res


class YaraProcessScanMatch(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.YaraProcessScanMatch
  rdf_deps = [rdf_client.Process, YaraMatch]


class YaraProcessScanMiss(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.YaraProcessScanMiss
  rdf_deps = [rdf_client.Process]


class YaraProcessScanResponse(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.YaraProcessScanResponse
  rdf_deps = [YaraProcessScanMatch, YaraProcessScanMiss, ProcessMemoryError]


class YaraProcessDumpArgs(rdf_structs.RDFProtoStruct):
  """Args for DumpProcessMemory flow and YaraProcessDump client action."""

  protobuf = flows_pb2.YaraProcessDumpArgs
  rdf_deps = [rdfvalue.ByteSize]

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    # These default values were migrated from the Protobuf definition.
    if not self.HasField("ignore_grr_process"):
      self.ignore_grr_process = True
    if not self.HasField("dump_all_processes"):
      self.dump_all_processes = False
    if not self.HasField("size_limit"):
      self.size_limit = 0
    if not self.HasField("chunk_size"):
      self.chunk_size = 100 * 1024 * 1024  # 100 MiB
    if not self.HasField("skip_special_regions"):
      self.skip_special_regions = False
    if not self.HasField("skip_mapped_files"):
      self.skip_mapped_files = True
    if not self.HasField("skip_shared_regions"):
      self.skip_shared_regions = False
    if not self.HasField("skip_executable_regions"):
      self.skip_executable_regions = False
    if not self.HasField("skip_readonly_regions"):
      self.skip_readonly_regions = False


class ProcessMemoryRegion(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.ProcessMemoryRegion
  rdf_deps = [rdf_paths.PathSpec]


class YaraProcessDumpInformation(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.YaraProcessDumpInformation
  rdf_deps = [rdf_client.Process, ProcessMemoryRegion]


class YaraProcessDumpResponse(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.YaraProcessDumpResponse
  rdf_deps = [YaraProcessDumpInformation, ProcessMemoryError]



#!/usr/bin/env python
"""A module with utilities for AEAD streams."""

from collections.abc import Iterator
import io
import itertools
import os
import struct
from typing import IO

from cryptography.hazmat.primitives.ciphers import aead

from grr_response_core.lib.util import io as ioutil
from grr_response_core.lib.util import iterator


def Encrypt(stream: IO[bytes], key: bytes) -> IO[bytes]:
  """Encrypts given file-like object using AES algorithm with GCM mode.

  The input stream is divided into small chunks of predefined size and then each
  chunk is encrypted using AES GCM procedure. In the encoded stream, before each
  proper chunk there is a nonce binary string prepended. As associated data for
  each encrypted chunk, chunk index and information about whether it is the last
  chunk is used.

  Args:
    stream: A file-like object to encrypt.
    key: A secret key used for encrypting the data.

  Returns:
    A file-like object with encrypted data.
  """
  aesgcm = aead.AESGCM(key)

  def Generate() -> Iterator[bytes]:
    chunks = ioutil.Chunk(stream, size=_AEAD_CHUNK_SIZE)
    chunks = iterator.Lookahead(enumerate(chunks))

    for idx, chunk in chunks:
      nonce = os.urandom(_AEAD_NONCE_SIZE)
      adata = _AEAD_ADATA_FORMAT.pack(idx, chunks.done)

      yield nonce + aesgcm.encrypt(nonce, chunk, adata)

  return ioutil.Unchunk(Generate())


def Decrypt(stream: IO[bytes], key: bytes) -> IO[bytes]:
  """Decrypts given file-like object using AES algorithm in GCM mode.

  Refer to the encryption documentation to learn about the details of the format
  that this function allows to decode.

  Args:
    stream: A file-like object to decrypt.
    key: A secret key used for decrypting the data.

  Returns:
    A file-like object with decrypted data.
  """
  aesgcm = aead.AESGCM(key)

  def Generate() -> Iterator[bytes]:
    # Buffered reader should accept `IO[bytes]` but for now it accepts only
    # `RawIOBase` (which is a concrete base class for all I/O implementations).
    reader = io.BufferedReader(stream)  # pytype: disable=wrong-arg-types

    # We abort early if there is no data in the stream. Otherwise we would try
    # to read nonce and fail.
    if not reader.peek():
      return

    for idx in itertools.count():
      nonce = reader.read(_AEAD_NONCE_SIZE)

      # As long there is some data in the buffer (and there should be because of
      # the initial check) there should be a fixed-size nonce prepended to each
      # chunk.
      if len(nonce) != _AEAD_NONCE_SIZE:
        raise EOFError(f"Incorrect nonce length: {len(nonce)}")

      chunk = reader.read(_AEAD_CHUNK_SIZE + 16)

      # `BufferedReader#peek` will return non-empty byte string if there is more
      # data available in the stream.
      is_last = reader.peek() == b""  # pylint: disable=g-explicit-bool-comparison

      adata = _AEAD_ADATA_FORMAT.pack(idx, is_last)

      yield aesgcm.decrypt(nonce, chunk, adata)

      if is_last:
        break

  return ioutil.Unchunk(Generate())


# We use 12 bytes (96 bits) as it is the recommended IV length by NIST for best
# performance [1]. See AESGCM documentation for more details.
#
# [1]: https://csrc.nist.gov/publications/detail/sp/800-38d/final
_AEAD_NONCE_SIZE = 12

# Because chunk size is crucial to the security of the whole procedure, we don't
# let users pick their own chunk size. Instead, we use a fixed-size chunks of
# 4 mebibytes.
_AEAD_CHUNK_SIZE = 4 * 1024 * 1024

# As associated data for each encrypted chunk we use an integer denoting chunk
# id followed by a byte with information whether this is the last chunk.
_AEAD_ADATA_FORMAT = struct.Struct("!Q?")



#!/usr/bin/env python
"""Configuration parameters for the server side subsystems."""


from grr_response_core import version
from grr_response_core.lib import config_lib
from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import paths as rdf_paths

VERSION = version.Version()

config_lib.DEFINE_integer("Source.version_major", VERSION["major"],
                          "Major version number of client binary.")

config_lib.DEFINE_integer("Source.version_minor", VERSION["minor"],
                          "Minor version number of client binary.")

config_lib.DEFINE_integer("Source.version_revision", VERSION["revision"],
                          "Revision number of client binary.")

config_lib.DEFINE_integer("Source.version_release", VERSION["release"],
                          "Release number of client binary.")

config_lib.DEFINE_string(
    "Source.version_string", "%(version_major).%(version_minor)."
    "%(version_revision).%(version_release)", "Version string of the client.")

config_lib.DEFINE_integer(
    "Source.version_numeric", "%(version_major)%(version_minor)"
    "%(version_revision)%(version_release)",
    "Version string of the client as an integer.")

# Note: Each thread adds about 8mb for stack space.
config_lib.DEFINE_integer("Threadpool.size", 50,
                          "Number of threads in the shared thread pool.")

config_lib.DEFINE_integer(
    "Worker.queue_shards", 5, "Queue notifications will be sharded across "
    "this number of datastore subjects.")

config_lib.DEFINE_list("Frontend.well_known_flows", [], "Unused, Deprecated.")

# Smtp settings.
config_lib.DEFINE_string("Worker.smtp_server", "localhost",
                         "The smtp server for sending email alerts.")

config_lib.DEFINE_integer("Worker.smtp_port", 25, "The smtp server port.")

config_lib.DEFINE_bool("Worker.smtp_starttls", False,
                       "Enable TLS for the smtp connection.")

config_lib.DEFINE_string("Worker.smtp_user", None,
                         "Username for the smtp connection.")

config_lib.DEFINE_string("Worker.smtp_password", None,
                         "Password for the smtp connection.")

config_lib.DEFINE_integer("Server.rsa_key_length", 2048,
                          "The length of the server rsa key in bits.")

config_lib.DEFINE_bool("Cron.active", False,
                       "Set to true to run a cron thread on this binary.")

config_lib.DEFINE_list(
    "Cron.disabled_cron_jobs",
    [],
    "Cron jobs listed here will not be run. Many system jobs are "
    "important. Leave empty unless you are sure that you "
    "know what you are doing.",
)

config_lib.DEFINE_integer(
    "Cron.interrogate_crash_limit", 500,
    "Maximum number of client crashes to allow for an Interrogate cron hunt "
    "before stopping the hunt.")

config_lib.DEFINE_integer(
    "Cron.interrogate_client_rate", 50,
    "Client rate setting for the periodical Interrogate cron hunt.")

config_lib.DEFINE_semantic_value(
    rdfvalue.Duration, "Cron.interrogate_duration",
    rdfvalue.Duration.From(1, rdfvalue.WEEKS),
    "Duration of the Interrogate cron hunt. The hunt is run weekly, so "
    "default duration is 1w. In certain cases the duration might be extended "
    "to accommodate for the clients that rarely show up online.")

config_lib.DEFINE_string("Frontend.bind_address", "::",
                         "The ip address to bind.")

config_lib.DEFINE_integer(
    "Frontend.max_queue_size", 500,
    "Maximum number of messages to queue for the client.")

config_lib.DEFINE_integer(
    "Frontend.max_retransmission_time", 10,
    "Maximum number of times we are allowed to "
    "retransmit a request until it fails.")

config_lib.DEFINE_integer(
    "Frontend.message_expiry_time", 600,
    "Maximum time messages remain valid within the "
    "system.")

config_lib.DEFINE_bool(
    "Server.initialized", False, "True once config_updater initialize has been "
    "run at least once.")

config_lib.DEFINE_string("Server.ip_resolver_class", "IPResolver",
                         "The ip resolver class to use.")

config_lib.DEFINE_string("Server.email_alerter_class", "SMTPEmailAlerter",
                         "The email alerter class to use.")

config_lib.DEFINE_string(
    "Server.username", None,
    "System account for services to run as after initialization. Note that "
    "GRR must be running as root first before being able to switch to another "
    "username. You would normally only need this if you want to bind to a low "
    "port for some reason.")

# Email Template Values
config_lib.DEFINE_string("Email.signature", "The GRR Team",
                         "The default signature block for template emails")

config_lib.DEFINE_string(
    "Email.approval_cc_address", None,
    "A single email address or comma separated list of addresses to CC on all "
    "approval emails. Will be added"
    " to all emails and can't be changed or removed by the user.")

config_lib.DEFINE_boolean(
    "Email.send_approval_emails", True,
    "Approval emails are sent for approvals in addition to notifications "
    "in the web UI.")

config_lib.DEFINE_string(
    "Email.approval_optional_cc_address", None,
    "A single email address or comma separated list of addresses to CC on all "
    "approval emails. The user has the option to"
    " remove this CC address .")

config_lib.DEFINE_string(
    "Email.approval_signature", None,
    "If you feel like it, you can add a funny cat picture to approval mails. "
    "Needs full html: <img src=\"https://imgur.com/path/to/cat.jpg\">.")

config_lib.DEFINE_bool(
    "Email.enable_custom_email_address", False,
    "If true, it's possible to set a custom E-Mail address for GRR users, "
    "overriding the default <username>@<Logging.domain>.")

config_lib.DEFINE_string(
    "StatsStore.process_id", default="", help="Unused, Deprecated.")

config_lib.DEFINE_integer(
    "StatsStore.write_interval", default=60, help="Unused, Deprecated")

config_lib.DEFINE_integer(
    "StatsStore.stats_ttl_hours", default=72, help="Unused, Deprecated.")

config_lib.DEFINE_bool(
    "AdminUI.allow_hunt_results_delete",
    default=False,
    help="If True, hunts with results can be deleted "
    "when the delete hunt button is used. Enable with "
    "caution as this allows erasure of historical usage for"
    "accountability purposes.")

config_lib.DEFINE_integer(
    "Server.max_unbound_read_size",
    10000000,
    help="The number of bytes allowed for unbounded reads from a file object")

# Data retention policies.
config_lib.DEFINE_semantic_value(
    rdfvalue.Duration,
    "DataRetention.cron_jobs_flows_ttl",
    default=None,
    help="Cron job flows TTL specified as the duration string. "
    "Examples: 90d, 180d, 1y. If not set, cron jobs flows will be retained "
    "forever.")

config_lib.DEFINE_semantic_value(
    rdfvalue.Duration,
    "DataRetention.hunts_ttl",
    default=None,
    help="Hunts TTL specified as the duration string. Examples: 90d, "
    "180d, 1y. If not set, hunts will be retained forever.")

config_lib.DEFINE_string(
    "DataRetention.hunts_ttl_exception_label",
    default="retain",
    help="Hunts marked with this label "
    "will be retained forever.")

config_lib.DEFINE_semantic_value(
    rdfvalue.Duration,
    "DataRetention.tmp_ttl",
    default=None,
    help="Temp TTL specified as the duration string. Examples: 90d, "
    "180d, 1y. If not set, temp objects will be retained forever.")

config_lib.DEFINE_string(
    "DataRetention.tmp_ttl_exception_label",
    default="retain",
    help="Temp objects marked with this "
    "label will be retained forever.")

config_lib.DEFINE_semantic_value(
    rdfvalue.Duration,
    "DataRetention.inactive_client_ttl",
    default=None,
    help="Temp TTL specified as the duration string. Examples: 90d, "
    "180d, 1y. If not set, inactive clients will be retained forever.")

config_lib.DEFINE_string(
    "DataRetention.inactive_client_ttl_exception_label",
    default="retain",
    help="Inactive clients marked with "
    "this label will be retained forever.")

config_lib.DEFINE_float(
    "Hunt.default_client_rate",
    default=20.0,
    help="Default value for HuntRunnerArgs.client_rate. Client rate "
    "determines how many clients per minute get to process a hunt")

config_lib.DEFINE_integer(
    "Hunt.default_crash_limit",
    default=100,
    help="Default value for HuntRunnerArgs.crash_limit. crash_limit is a "
    "maximum number of clients that are allowed to crash before the hunt is "
    "automatically hard-stopped.")

config_lib.DEFINE_integer(
    "Hunt.default_avg_results_per_client_limit",
    default=1000,
    help="Default value for HuntRunnerArgs.avg_results_per_client_limit. "
    "If the average number of results per client is greater than "
    "avg_results_per_client_limit, the hunt gets stopped.")

config_lib.DEFINE_integer(
    "Hunt.default_avg_cpu_seconds_per_client_limit",
    default=60,
    help="Default value for HuntRunnerArgs.avg_cpu_seconds_per_client_limit. "
    "If the average CPU usage seconds per client becomes "
    "greater than this limit, the hunt gets stopped.")

config_lib.DEFINE_integer(
    "Hunt.default_avg_network_bytes_per_client_limit",
    default=10 * 1024 * 1024,  # 10Mb
    help="Default value for HuntRunnerArgs.avg_network_bytes_per_client_limit. "
    "If the average network usage per client becomes "
    "greater than this limit, the hunt gets stopped.")

# GRRafana HTTP Server settings.
config_lib.DEFINE_string(
    "GRRafana.bind", default="localhost", help="The GRRafana server address.")

config_lib.DEFINE_integer(
    "GRRafana.port", default=5000, help="The GRRafana server port.")

# Fleetspeak server-side integration flags.
config_lib.DEFINE_string(
    "Server.fleetspeak_message_listen_address", "",
    "The Fleetspeak server message listen address, formatted like "
    "localhost:6061.")

config_lib.DEFINE_string(
    "Server.fleetspeak_server", "",
    "The Fleetspeak server address, formatted like localhost:6062.")

config_lib.DEFINE_string(
    "Server.fleetspeak_label_prefix", None,
    "Prefix used to identify Fleetspeak labels that should be written to "
    "GRR's DB during Interrogation. If not given, all labels are written.")

config_lib.DEFINE_list(
    "Server.fleetspeak_label_map", [],
    "Used to override fleetspeak labels with custom labels. Entries in the "
    "list are expected to be of the form '<fleetspeak-label>:<override>'. If "
    "a Fleetspeak label is not in the map, it will be written as is to GRR's "
    "DB as part of the Interrogate flow.")

config_lib.DEFINE_semantic_value(
    rdfvalue.Duration,
    "Server.fleetspeak_last_ping_threshold",
    default="2h",
    help="Age above which to consider last-ping timestamps for Fleetspeak "
    "clients as stale, and in need of updating (by querying Fleetspeak "
    "servers).")

config_lib.DEFINE_integer(
    "Server.fleetspeak_list_clients_batch_size",
    default=5000,
    help="Maximum number of client ids to place in a single Fleetspeak "
    "ListClients() API request.")

config_lib.DEFINE_bool(
    "Server.fleetspeak_cps_enabled",
    default=False,
    help=(
        "Enable the Google Cloud Pub/Sub communication channel from Fleetspeak."
        " When this is set to True, incoming messages from Fleetspeak are"
        " fetched from a Cloud Pub/Sub subscription instead of being received"
        " directly via GRPC."
    ),
)

config_lib.DEFINE_string(
    "Server.fleetspeak_cps_project",
    default=None,
    help=(
        "Google Cloud project name to use when communicating with Fleetspeak "
        "via Cloud Pub/Sub."
    ),
)

config_lib.DEFINE_string(
    "Server.fleetspeak_cps_subscription",
    default=None,
    help=(
        "Cloud Pub/Sub subscription name to use for reading Fleetspeak "
        "messages. This subscription must have been already created."
    ),
)

config_lib.DEFINE_integer(
    "Server.fleetspeak_cps_concurrency",
    default=20,
    help=(
        "The number of concurrent message-processing subscribers to spawn "
        "when receiving Fleetspeak messages via Cloud Pub/Sub."
    ),
)

config_lib.DEFINE_semantic_enum(
    rdf_paths.PathSpec.PathType,
    "Server.raw_filesystem_access_pathtype",
    default=rdf_paths.PathSpec.PathType.NTFS,
    help="PathType to use for raw filesystem access on Windows.")

config_lib.DEFINE_boolean(
    "Server.grr_binaries_readonly", False,
    "When set to True, uploaded GRR binaries can't be deleted or overwritten.")

config_lib.DEFINE_boolean(
    name="Interrogate.collect_crowdstrike_agent_id",
    default=False,
    help=(
        "Whether the interrogate flow should collect identifier of the "
        "endpoint's CrowdStrike agent."
    ),
)

config_lib.DEFINE_boolean(
    name="Interrogate.collect_passwd_cache_users",
    default=False,
    help=(
        "Whether the interrogate flow should collect user information using the"
        "`/etc/passwd.cache` file."
    ),
)

config_lib.DEFINE_string(
    "Server.signed_url_service_account_email",
    default=None,
    help=(
        "The email of the Service Account to use for signing the URL"
        " (https://cloud.google.com/storage/docs/access-control/signed-urls#signing-resumable)."
    ),
)

config_lib.DEFINE_string(
    "Server.signed_url_gcs_bucket_name",
    default=None,
    help=(
        "The GCS bucket name to include in the signed URL"
        " (https://cloud.google.com/storage/docs/access-control/signed-urls#signing-resumable)."
    ),
)

config_lib.DEFINE_integer(
    "Server.signed_url_expire_hours",
    default=12,
    help=(
        "The TTL until the signed URL expires"
        " (https://cloud.google.com/storage/docs/access-control/signed-urls#signing-resumable)."
    ),
)

config_lib.DEFINE_string(
    "Server.disable_rrg_support",
    default=False,
    help=(
        "Disables support for RRG agents (forces the traffic to always be "
        "routed to the Python agent)."
    ),
)

config_lib.DEFINE_string(
    name="CommandSigning.ed25519_private_key_file",
    default=None,
    help="An UTF-8 encoded Ed25519 encryption key to sign commands.",
)
